import jax
import jax.numpy as jnp
from .abstract import *
from ..utils import *
from .truncatedgaussian1D import logC
import pandas as pd
import numpy as np

class Uniform1DAnalytic(AnalyticPopulationModel):
    def __init__(self, a, b, var_names=['chi_1'], hyper_var_names=[]):
        self.a = a
        self.b = b
        self.norm = (self.b - self.a)
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}
    
    def get_data(self, data, params, component):
        var_name = self.var_name_1 if (component == 1) else self.var_name_2
        X_locations = data[var_name + '_mu_kernel'];
        X_scales    = data[var_name + '_sigma_kernel'];
        return X_locations, X_scales
    
    def evaluate(self, data, params):
        x_1 = data[self.var_names[0]]
        return jnp.ones(len(x_1)) / self.norm
        
    def __call__(self, data, params):
        X_locations = data[self.var_names[0] + '_mu_kernel'];
        return jnp.ones_like(X_locations) / self.norm
    
    def _sample(self, df_hyper_samples):
        samps = {}
        for var in self.var_names:
            a,b = self.limits[var]
            samps[var] = a + np.random.rand(len(df_hyper_samples))*(b-a)
        return pd.DataFrame(samps)

    def sample(self, df_hyper_samples, oversample=1):
        return pd.concat([self._sample(df_hyper_samples) for _ in range(oversample)])


class Uniform2DAnalytic(AnalyticPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=[]):
        self.a = a
        self.b = b
        self.norm = np.prod([(self.b[i] - self.a[i]) for i in range(len(a))])
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.var_name_1 = var_names[0]
        self.var_name_2 = var_names[1]

    @property
    def limits(self):
        return {var : [self.a[i], self.b[i]] for i,var in enumerate(self.var_names)}
    
    def get_data(self, data, params, component):
        var_name = self.var_name_1 if (component == 1) else self.var_name_2
        X_locations = data[var_name + '_mu_kernel'];
        X_scales    = data[var_name + '_sigma_kernel'];
        return X_locations, X_scales
    
    def evaluate(self, data, params):
        x_1 = data[self.var_name_1]
        return jnp.ones(len(x_1)) / self.norm
        
    def __call__(self, data, params):
        X_locations, X_scales = self.get_data(data, params, component=1);
        return jnp.ones_like(X_locations) / self.norm
    
    def _sample(self, df_hyper_samples):
        samps = {}
        for var in self.var_names:
            a,b = self.limits[var]
            samps[var] = a + np.random.rand(len(df_hyper_samples))*(b-a)
        return pd.DataFrame(samps)

    def sample(self, df_hyper_samples, oversample=1):
        return pd.concat([self._sample(df_hyper_samples) for _ in range(oversample)])



class Uniform1DAnalyticVariedLimits(AnalyticPopulationModel):
    def __init__(self, a, b, var_names=['x'], hyper_var_names=['x_min', 'x_max']):
        self.a = a
        self.b = b
        self.var_name = var_names[0]
        self.x_min = hyper_var_names[0];
        self.x_max = hyper_var_names[1];
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}
        
    def get_data(self, data, params):
        X_locations = data[self.var_name + '_mu_kernel'];
        X_scales    = data[self.var_name + '_sigma_kernel'];
        x_min       = params.get(self.x_min, self.a);
        x_max       = params.get(self.x_max, self.b); 
        return X_locations, X_scales, x_min, x_max

    def evaluate(self, data, params):
        """
        Evaluates the value of the *integrand* and not the integral
        """
        x_min       = params.get(self.x_min, self.a);
        x_max       = params.get(self.x_max, self.b);
        #loglikes = jax.scipy.special.logsumexp( jnp.log(truncnorm(Xs, mu, sigma, low=a, high=b)) , axis=-1) - jnp.log(Xs.shape[-1])
        prob = box(data[self.var_name], x_min,x_max)/(x_max - x_min)
        return prob
        
    def __call__(self, data, params):
        X_locations, X_scales, x_min, x_max = self.get_data(data, params);
        X_stddevs_outside = jnp.where(X_locations <= x_max, 
                                        jnp.where(X_locations <= x_min, 
                                            (x_min - X_locations)/X_scales, jnp.zeros_like(X_locations)),
                                                    (X_locations - x_max)/X_scales)
        loglikes = logC(X_locations, X_scales, x_min, x_max) - logC(X_locations, X_scales, self.a, self.b) - jnp.log(x_max - x_min)
        likes = jnp.where(X_stddevs_outside < 4, jnp.exp(loglikes), jnp.zeros_like(loglikes))
        return likes

    def sample(self, df_hyper_samples, oversample=1):
        N = len(df_hyper_samples)
        u = np.random.rand(N);
        x_max = df_hyper_samples[self.x_max];
        x_min = df_hyper_samples[self.x_min];
        
        return pd.DataFrame({self.var_name : u*(x_max - x_min) + x_min})


class Uniform1DIndependentAnalyticVariedLimits(AnalyticPopulationModel):
    def __init__(self, a, b, var_names=['x', 'y'], hyper_var_names=['x_min', 'x_max', 'y_min', 'y_max']):
        self.a, self.b = a,b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        kwargs = {'a' : a, 'b' : b}
        self.models = [Uniform1DAnalyticVariedLimits(var_names=[var_names[i]], 
                                                   hyper_var_names=[hyper_var_names[2*i], hyper_var_names[2*i + 1]], 
                                                   **kwargs) for i in range(len(self.var_names))]
    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def evaluate(self, data, params):
        """
        Evaluates the value of the *integrand* and not the integral
        """
        result = self.models[0].evaluate(data, params)
        for i in range(1,len(self.models)):
            result *= self.models[i].evaluate(data, params)
        return result
        
    def __call__(self, data, params):
        result = self.models[0](data, params)
        for i in range(1,len(self.models)):
            result *= self.models[i](data, params)
        return result

    def sample(self, df_hyper_samples, oversample=1):
        return pd.concat([model.sample(df_hyper_samples, oversample=oversample) for model in self.models])


class Uniform2DAnalyticVariedLimits(AnalyticPopulationModel):
    def __init__(self, a, b, var_names=['x', 'y'], hyper_var_names=['x_min', 'x_max', 'y_min', 'y_max']):
        self.a = a
        self.b = b
        self.x_var_name = var_names[0]
        self.y_var_name = var_names[1]
        self.x_min = hyper_var_names[0];
        self.x_max = hyper_var_names[1];
        self.y_min = hyper_var_names[2];
        self.y_max = hyper_var_names[3];
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names

    @property
    def limits(self):
        return {var : [self.a[i], self.b[i]] for i,var in enumerate(self.var_names)}
        
    def get_data(self, data, params):
        X_locations = data[self.x_var_name + '_mu_kernel'];
        X_scales    = data[self.x_var_name + '_sigma_kernel'];
        Y_locations = data[self.y_var_name + '_mu_kernel'];
        Y_scales    = data[self.y_var_name + '_sigma_kernel'];
        XY_rho      = data.get(self.x_var_name + '_rho_kernel', 1e-6);
        x_min       = params.get(self.x_min, self.a[0])
        x_max       = params.get(self.x_max, self.b[0])
        y_min       = params.get(self.y_min, self.a[1])
        y_max       = params.get(self.y_max, self.b[1])
        return X_locations, X_scales, x_min, x_max, Y_locations, Y_scales, y_min, y_max, XY_rho

    def evaluate(self, data, params):
        """
        Evaluates the value of the *integrand* and not the integral
        """
        x_min       = params.get(self.x_min, self.a[0])
        x_max       = params.get(self.x_max, self.b[0])
        y_min       = params.get(self.y_min, self.a[1])
        y_max       = params.get(self.y_max, self.b[1])
        #loglikes = jax.scipy.special.logsumexp( jnp.log(truncnorm(Xs, mu, sigma, low=a, high=b)) , axis=-1) - jnp.log(Xs.shape[-1])
        prob = box(data[self.x_var_name], x_min,x_max) * box(data[self.y_var_name], y_min,y_max) /((x_max - x_min)*(y_max - y_min))
        return prob
        
    def __call__(self, data, params):
        X_locations, X_scales, x_min, x_max, Y_locations, Y_scales, y_min, y_max, XY_rho = self.get_data(data, params);
        X_stddevs_outside = jnp.where(X_locations <= x_max, 
                                        jnp.where(X_locations <= x_min, 
                                            (x_min - X_locations)/X_scales, jnp.zeros_like(X_locations)),
                                                    (X_locations - x_max)/X_scales)
        Y_stddevs_outside = jnp.where(Y_locations <= y_max, 
                                        jnp.where(Y_locations <= y_min, 
                                            (y_min - Y_locations)/Y_scales, jnp.zeros_like(Y_locations)),
                                                    (Y_locations - y_max)/Y_scales)
        overlap = mvnorm2d(X_locations, Y_locations, X_scales, Y_scales, x_min, y_min, x_max, y_max, XY_rho)
        overlap /= (x_max - x_min)*(y_max - y_min)
        overlap /= mvnorm2d(X_locations, Y_locations, X_scales, Y_scales, self.a[0], self.a[1], self.b[0], self.b[1], XY_rho)
        new_overlap = jnp.where(X_stddevs_outside < 4, jnp.where(Y_stddevs_outside < 4, overlap, jnp.zeros_like(overlap)),jnp.zeros_like(overlap))
        return new_overlap

    def sample(self, df_hyper_samples, oversample=1):
        N = len(df_hyper_samples)
        u1 = np.random.rand(N);
        u2 = np.random.rand(N);
        x_min       = params.get(self.x_min, self.a[0])
        x_max       = params.get(self.x_max, self.b[0])
        y_min       = params.get(self.y_min, self.a[1])
        y_max       = params.get(self.y_max, self.b[1])
        return pd.DataFrame({self.x_var_name : u1*(x_max - x_min) + x_min, self.y_var_name : u2*(y_max - y_min) + y_min})





import jax.numpy as jnp
from jax import grad

class BetaDistributionConverter:
    def __init__(self):
        pass
    
    @staticmethod
    def _convert_to_beta_parameters(parameters):
        added_keys = []
        converted = parameters.copy()

        def _convert(suffix):
            alpha = f"alpha_chi{suffix}"
            beta = f"beta_chi{suffix}"
            mu = f"mu_chi{suffix}"
            sigma = f"sigma_chi{suffix}"
            amax = f"amax{suffix}"

            if alpha not in parameters or beta not in parameters:
                needed = True
            elif converted[alpha] is None or converted[beta] is None:
                needed = True
            else:
                needed = False
                done = True

            if needed:
                if mu in converted and sigma in converted:
                    done = True
                    alpha_val, beta_val, _ = BetaDistributionConverter.mu_var_max_to_alpha_beta_max(
                        converted[mu], converted[sigma], converted[amax]
                    )
                    converted[alpha], converted[beta] = alpha_val, beta_val
                    added_keys.extend([alpha, beta])
                else:
                    done = False
            return done

        done = False
        for suffix in ["_1", "_2", ""]:
            done |= _convert(suffix)

        return converted, added_keys

    @staticmethod
    def alpha_beta_max_to_mu_var_max(alpha, beta, amax):
        mu = alpha / (alpha + beta) * amax
        var = alpha * beta / ((alpha + beta) ** 2 * (alpha + beta + 1)) * amax ** 2
        return mu, var, amax

    @staticmethod
    def mu_var_max_to_alpha_beta_max(mu, var, amax):
        mu /= amax
        var /= amax ** 2
        alpha = (mu ** 2 * (1 - mu) - mu * var) / var
        beta = (mu * (1 - mu) ** 2 - (1 - mu) * var) / var
        return alpha, beta, amax

    def convert_parameters(self, parameters_dict, remove=True):
        converted_dict, added_keys = self._convert_to_beta_parameters(parameters_dict)
        if remove:
            for key in added_keys:
                converted_dict.pop(key, None)
        return converted_dict


import jax
import jax.numpy as jnp
from ..utils import *
from ..generic import *
from ..sample import *


class GaussianIsotropicSpinOrientationsIIDAnalytic(AnalyticPopulationModel, SpinPopulationModel):
    r"""
    Mixture of gaussian and isotropic distribution over spin orientations.
    Performs a monte carlo estimate of the population likelihood. 

    Event Level Parameters:         :math:`z_1=cos(\theta_1), z_2=cos(\theta_2)`
    Population Level Parameters:    :math:`\xi, \sigma` 

    .. math::
    
        P(z_1, z_2| \xi, \sigma) = \frac{1-\xi}{4} + \xi \left(  \mathcal{N}_{[-1,1]}(z_1 | \xi, \sigma) \mathcal{N}_{[-1,1]}(z_2 | \xi, \sigma) \right) 
    """
    def __init__(self, var_names=['cos_tilt_1', 'cos_tilt_2'], hyper_var_names=['xi_spin','sigma_spin'], a=-1, b=1):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.models = [TruncatedGaussian1DAnalytic(a=a, b=b, var_names=[var_names[i]], hyper_var_names=['mu_spin', hyper_var_names[1]]) for i in range(len(var_names))]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def evaluate(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        sigma_spin = params[self.hyper_var_names[1]]
        prob  = truncnorm(data[self.var_names[0]], mu=1, sigma=sigma_spin, high=self.b, low=self.a)
        prob *= truncnorm(data[self.var_names[1]], mu=1, sigma=sigma_spin, high=self.b, low=self.a)
        prob *= xi_spin
        prob += (1-xi_spin)/4
        return prob
    
    def __call__(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        sigma_spin = params[self.hyper_var_names[1]]
        new_params = {**params, 'mu_spin':1}
        prob  = self.models[0](data, new_params)           #truncnorm(data[self.var_names[0]], mu=1, sigma=sigma_spin, high=self.b, low=self.a)
        prob *= self.models[1](data, new_params)
        prob *= xi_spin
        prob += (1-xi_spin)/4
        return prob

class GaussianIsotropicSpinOrientationsFloatingAnalytic(AnalyticPopulationModel, SpinPopulationModel):
    r"""
    Mixture of gaussian and isotropic distribution over spin orientations.
    Performs a monte carlo estimate of the population likelihood. 

    Event Level Parameters:         :math:`z_1=cos(\theta_1), z_2=cos(\theta_2)`
    Population Level Parameters:    :math:`\xi, \sigma` 

    .. math::
    
        P(z_1, z_2| \xi, \sigma) = \frac{1-\xi}{4} + \xi \left(  \mathcal{N}_{[-1,1]}(z_1 | \xi, \sigma) \mathcal{N}_{[-1,1]}(z_2 | \xi, \sigma) \right) 
    """
    def __init__(self, var_names=['cos_tilt_1', 'cos_tilt_2'], hyper_var_names=['xi_spin','mu_spin_1', 'sigma_spin_1', 'mu_spin_2', 'sigma_spin_2'], a=-1, b=1):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.models = [TruncatedGaussian1DAnalytic(a=a, b=b, var_names=[var_names[i]], hyper_var_names=[hyper_var_names[2*i + 1], hyper_var_names[2*i + 2]]) for i in range(len(var_names))]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def evaluate(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        mu_spin_1, mu_spin_2 = params[self.hyper_var_names[1]], params[self.hyper_var_names[3]]
        sigma_spin_1, sigma_spin_2 = params[self.hyper_var_names[2]], params[self.hyper_var_names[4]]
        prob  = truncnorm(data[self.var_names[0]], mu=mu_spin_1, sigma=sigma_spin_1, high=self.b, low=self.a)
        prob *= truncnorm(data[self.var_names[1]], mu=mu_spin_2, sigma=sigma_spin_2, high=self.b, low=self.a)
        prob *= xi_spin
        prob += (1-xi_spin)/4
        return prob
    
    def __call__(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        mu_spin_1, mu_spin_2 = params[self.hyper_var_names[1]], params[self.hyper_var_names[3]]
        sigma_spin_1, sigma_spin_2 = params[self.hyper_var_names[2]], params[self.hyper_var_names[4]]
        #new_params = {**params, 'mu_spin':1}
        prob  = self.models[0](data, new_params)           #truncnorm(data[self.var_names[0]], mu=1, sigma=sigma_spin, high=self.b, low=self.a)
        prob *= self.models[1](data, new_params)
        prob *= xi_spin
        prob += (1-xi_spin)/4
        return prob



class IIDTruncatedGaussian1DAnalytic(AnalyticPopulationModel):
    def __init__(self, a, b, var_names=['x'], hyper_var_names=['mu', 'sigma']):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        kwargs = {'a' : a, 'b' : b, 'hyper_var_names' : hyper_var_names}
        self.models = [TruncatedGaussian1DAnalytic(var_names=[var_name], **kwargs) for var_name in self.var_names]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        result = self.models[0](data, params)
        for i in range(1,len(self.models)):
            result *= self.models[i](data, params)
        return result

    def evaluate(self, data, params):
        result = self.models[0].evaluate(data, params)
        for i in range(1,len(self.models)):
            result *= self.models[i].evaluate(data, params)
        return result


class MixtureTruncated1DAnalyticFixed0(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['x'], hyper_var_names=['sigma_0', 'mu', 'sigma', 'eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.eta_variable = hyper_var_names[-1]
        kwargs = {'a' : a, 'b' : b, 'hyper_var_names' : hyper_var_names}

        comp1_hyper_var_names = ['mu_zero_spin_1d_fixed', self.hyper_var_names[0]]
        comp2_hyper_var_names = [self.hyper_var_names[1], self.hyper_var_names[2]]

        self.models = [TruncatedGaussian1DAnalytic(var_names=var_names, hyper_var_names=comp1_hyper_var_names, a=a, b=b),
                       TruncatedGaussian1DAnalytic(var_names=var_names, hyper_var_names=comp2_hyper_var_names, a=a, b=b)]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        eta = params[self.eta_variable]
        result  = eta * self.models[0](data, params)
        result += (1-eta) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        eta = params[self.eta_variable]
        result  = eta * self.models[0].evaluate(data, params)
        result += (1-eta) * self.models[1].evaluate(data, params)
        return result

class IIDTruncatedGaussian1DAnalyticMixture(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['x'], hyper_var_names=['sigma_0', 'mu', 'sigma', 'eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        kwargs = {'a' : a, 'b' : b, 'hyper_var_names' : hyper_var_names}
        self.models = [MixtureTruncated1DAnalyticFixed0(var_names=[var_name], **kwargs) for var_name in self.var_names]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        result = self.models[0](data, params)
        for i in range(1,len(self.models)):
            result *= self.models[i](data, params)
        return result

    def evaluate(self, data, params):
        result = self.models[0].evaluate(data, params)
        for i in range(1,len(self.models)):
            result *= self.models[i].evaluate(data, params)
        return result


class TruncatedGaussian1DIndependentAnalytic(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['mu_chi_1', 'sigma_chi_1', 'mu_chi_2', 'sigma_chi_2']):
        self.a, self.b = a,b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        kwargs = {'a' : a, 'b' : b}
        self.models = [TruncatedGaussian1DAnalytic(var_names=[var_names[i]], 
                                                   hyper_var_names=[hyper_var_names[2*i], hyper_var_names[2*i + 1]], 
                                                   **kwargs) for i in range(len(self.var_names))]
    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        result = self.models[0](data, params)
        for i in range(1,len(self.models)):
            result *= self.models[i](data, params)
        return result

    def evaluate(self, data, params):
        result = self.models[0].evaluate(data, params)
        for i in range(1,len(self.models)):
            result *= self.models[i].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples, oversample=1):
        return ppd_truncUncorrelatedanalytic(self, df_hyper_samples, oversample=oversample)

class TruncatedGaussian1DIndependentAnalyticVariedLimits(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['mu_chi_1', 'sigma_chi_1', 'chi_1_min', 'chi_1_max', 
                                                                            'mu_chi_2', 'sigma_chi_2', 'chi_2_min', 'chi_2_max']):
        self.a, self.b = a,b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        kwargs = {'a' : a, 'b' : b}
        self.models = [TruncatedGaussian1DAnalyticVariedLimits(var_names=[var_names[i]], 
                                                   hyper_var_names=[hyper_var_names[4*i], hyper_var_names[4*i + 1], hyper_var_names[4*i + 2], hyper_var_names[4*i + 3]], 
                                                   **kwargs) for i in range(len(self.var_names))]
    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        result = self.models[0](data, params)
        for i in range(1,len(self.models)):
            result *= self.models[i](data, params)
        return result

    def evaluate(self, data, params):
        result = self.models[0].evaluate(data, params)
        for i in range(1,len(self.models)):
            result *= self.models[i].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples, oversample=1):
        return ppd_truncUncorrelatedanalytic(self, df_hyper_samples, oversample=oversample)


class TruncatedGaussian1DMixtureZeroFloatingAndUniform(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['mu_chi_1', 'sigma_chi_1', 'mu_chi_2', 'sigma_chi_2','eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.a = a; self.b=b;
        kwargs = {'a' : a, 'b' : b}
        comp1_var_names = [var_names[0]]
        comp2_var_names = [var_names[1]]

        comp1_hyper_var_names = [self.hyper_var_names[0], self.hyper_var_names[1], self.hyper_var_names[2], self.hyper_var_names[3]]
        comp2_hyper_var_names = []
        self.mixture_hyper_var_name = self.hyper_var_names[-1]

        self.models = [TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp1_hyper_var_names, a=a, b=b),
                       Uniform2DAnalytic(var_names=self.var_names, hyper_var_names=comp2_hyper_var_names, a=[a,a], b=[b,b])]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples_in, oversample=1, **kwargs):
        kwargs['oversample'] = oversample
        df_hyper_samples = df_hyper_samples_in.copy()
        series_1 = self.models[0].sample(df_hyper_samples, **kwargs)
        series_2 = self.models[1].sample(df_hyper_samples, **kwargs)
        N = len(df_hyper_samples)
        sampled_mixture = np.hstack([(df_hyper_samples[self.mixture_hyper_var_name] <= np.random.rand(N)) for _ in range(oversample)])
        d = {col : ((sampled_mixture * series_1[col]) + ((1 - sampled_mixture) * series_2[col])) for col in series_1.columns}
        return pd.DataFrame(d)


class TruncatedGaussian1DMixtureZeroAndFloatingIID(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['sigma_chi_at_0','mu_chi', 'sigma_chi', 'eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.a = a; self.b=b;
        kwargs = {'a' : a, 'b' : b}
        comp1_var_names = [var_names[0]]
        comp2_var_names = [var_names[1]]

        comp1_hyper_var_names = ['mu_zero_spin_1d_fixed', self.hyper_var_names[0], 'mu_zero_spin_1d_fixed', self.hyper_var_names[0]]
        comp2_hyper_var_names = self.hyper_var_names[1:3]*2
        self.mixture_hyper_var_name = self.hyper_var_names[-1]

        self.models = [TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp1_hyper_var_names, a=a, b=b),
                       TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp2_hyper_var_names, a=a, b=b)]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples_in, oversample=1, **kwargs):
        kwargs['oversample'] = oversample
        df_hyper_samples = df_hyper_samples_in.copy()
        df_hyper_samples['mu_zero_spin_1d_fixed'] = 0
        series_1 = self.models[0].sample(df_hyper_samples, **kwargs)
        series_2 = self.models[1].sample(df_hyper_samples, **kwargs)
        N = len(df_hyper_samples)
        sampled_mixture = np.hstack([(df_hyper_samples[self.mixture_hyper_var_name] <= np.random.rand(N)) for _ in range(oversample)])
        d = {col : ((sampled_mixture * series_1[col]) + ((1 - sampled_mixture) * series_2[col])) for col in series_1.columns}
        return pd.DataFrame(d)



class TruncatedGaussian2DMixtureZeroAndFloatingIID(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['sigma_chi_at_0','mu_chi', 'sigma_chi', 'eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.a = a; self.b=b;
        kwargs = {'a' : a, 'b' : b}
        comp1_var_names = [var_names[0]]
        comp2_var_names = [var_names[1]]

        comp1_hyper_var_names = ['mu_zero_spin_1d_fixed', self.hyper_var_names[0], 'mu_zero_spin_1d_fixed', self.hyper_var_names[0], 'rho_zero_fixed']
        comp2_hyper_var_names = self.hyper_var_names[1:3]*2 + ['rho_zero_fixed']
        self.mixture_hyper_var_name = self.hyper_var_names[-1]

        self.models = [TruncatedGaussian2DAnalytic(var_names=self.var_names, hyper_var_names=comp1_hyper_var_names, a=[a,a], b=[b,b]),
                       TruncatedGaussian2DAnalytic(var_names=self.var_names, hyper_var_names=comp2_hyper_var_names, a=[a,a], b=[b,b])]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        params['rho_zero_fixed'] = 1e-6    # Small non-zero number to prevent nan gradients
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        params['rho_zero_fixed'] = 1e-6    # Small non-zero number to prevent nan gradients
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples_in, oversample=1, **kwargs):
        kwargs['oversample'] = oversample
        df_hyper_samples = df_hyper_samples_in.copy()
        df_hyper_samples['mu_zero_spin_1d_fixed'] = 0
        df_hyper_samples['rho_zero_fixed'] = 0
        series_1 = self.models[0].sample(df_hyper_samples, **kwargs)
        series_2 = self.models[1].sample(df_hyper_samples, **kwargs)
        N = len(df_hyper_samples)
        sampled_mixture = np.hstack([(df_hyper_samples[self.mixture_hyper_var_name] <= np.random.rand(N)) for _ in range(oversample)])
        d = {col : ((sampled_mixture * series_1[col]) + ((1 - sampled_mixture) * series_2[col])) for col in series_1.columns}
        return pd.DataFrame(d)


class GaussianIsotropicSpinOrientationsIIDAnalytic2D(AnalyticPopulationModel, SpinPopulationModel):
    r"""
    Mixture of gaussian and isotropic distribution over spin orientations.
    Performs a monte carlo estimate of the population likelihood. 

    Event Level Parameters:         :math:`z_1=cos(\theta_1), z_2=cos(\theta_2)`
    Population Level Parameters:    :math:`\xi, \sigma` 

    .. math::
    
        P(z_1, z_2| \xi, \sigma) = \frac{1-\xi}{4} + \xi \left(  \mathcal{N}_{[-1,1]}(z_1 | \xi, \sigma) \mathcal{N}_{[-1,1]}(z_2 | \xi, \sigma) \right) 
    """
    def __init__(self, var_names=['cos_tilt_1', 'cos_tilt_2'], hyper_var_names=['xi_spin','sigma_spin'], a=-1, b=1):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.model = TruncatedGaussian2DAnalytic(a=[a,a], b=[b,b], var_names=var_names, hyper_var_names=['mu_spin', hyper_var_names[1], 'mu_spin', hyper_var_names[1], 'rho_zero_fixed'])

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def evaluate(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        sigma_spin = params[self.hyper_var_names[1]]
        prob  = truncnorm(data[self.var_names[0]], mu=1, sigma=sigma_spin, high=self.b, low=self.a)
        prob *= truncnorm(data[self.var_names[1]], mu=1, sigma=sigma_spin, high=self.b, low=self.a)
        prob *= xi_spin
        prob += (1-xi_spin)/4
        return prob
        
    def __call__(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        sigma_spin = params[self.hyper_var_names[1]]
        new_params = {**params, 'mu_spin':1, 'rho_zero_fixed':1e-6}
        prob  = self.model(data, new_params)
        prob *= xi_spin
        prob += (1-xi_spin)/4
        return prob


class GaussianIsotropicSpinOrientationsIIDAnalytic2DWithMinimum(AnalyticPopulationModel, SpinPopulationModel):
    r"""
    Mixture of gaussian and isotropic distribution over spin orientations.
    Performs a monte carlo estimate of the population likelihood. 

    Event Level Parameters:         :math:`z_1=cos(\theta_1), z_2=cos(\theta_2)`
    Population Level Parameters:    :math:`\xi, \sigma` 

    .. math::
    
        P(z_1, z_2| \xi, \sigma) = \frac{1-\xi}{4} + \xi \left(  \mathcal{N}_{[-1,1]}(z_1 | \xi, \sigma) \mathcal{N}_{[-1,1]}(z_2 | \xi, \sigma) \right) 
    """
    def __init__(self, var_names=['cos_tilt_1', 'cos_tilt_2'], hyper_var_names=['xi_spin','sigma_spin', 'z_min'], a=-1, b=1):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.model = TruncatedGaussian2DAnalyticLimits(a=[a,a], b=[b,b], 
                                                var_names=var_names, 
                                                hyper_var_names=['mu_spin', hyper_var_names[1], 
                                                                 'mu_spin', hyper_var_names[1], 'rho_zero_fixed',
                                                                 'z_min', 'z_max_fixed_to_b', 'z_min', 'z_max_fixed_to_b'])

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def evaluate(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        sigma_spin = params[self.hyper_var_names[1]]
        z_min = params[self.hyper_var_names[2]]
        prob  = truncnorm(data[self.var_names[0]], mu=1, sigma=sigma_spin, high=self.b, low=z_min)
        prob *= truncnorm(data[self.var_names[1]], mu=1, sigma=sigma_spin, high=self.b, low=z_min)
        prob *= xi_spin
        prob += (1-xi_spin)/((1-z_min)**2)
        return prob
        
    def __call__(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        sigma_spin = params[self.hyper_var_names[1]]
        z_min = params[self.hyper_var_names[2]]
        new_params = {**params, 'mu_spin':1, 'rho_zero_fixed':1e-6,  'z_max_fixed_to_b' : self.b}
        prob  = self.model(data, new_params)
        prob *= xi_spin
        prob += (1-xi_spin)/((1-z_min)**2)
        return prob

class GaussianIsotropicSpinOrientationsIIDAnalytic1DWithMinimum(AnalyticPopulationModel, SpinPopulationModel):
    r"""
    Mixture of gaussian and isotropic distribution over spin orientations.
    Performs a monte carlo estimate of the population likelihood. 

    Event Level Parameters:         :math:`z_1=cos(\theta_1), z_2=cos(\theta_2)`
    Population Level Parameters:    :math:`\xi, \sigma` 

    .. math::
    
        P(z_1, z_2| \xi, \sigma) = \frac{1-\xi}{4} + \xi \left(  \mathcal{N}_{[-1,1]}(z_1 | \xi, \sigma) \mathcal{N}_{[-1,1]}(z_2 | \xi, \sigma) \right) 
    """
    def __init__(self, var_names=['cos_tilt_1', 'cos_tilt_2'], hyper_var_names=['xi_spin','sigma_spin', 'z_min'], a=-1, b=1):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.models = [TruncatedGaussian1DAnalyticVariedLimits(a=a, b=b, var_names=[var_names[i]], hyper_var_names=['mu_spin', hyper_var_names[1], hyper_var_names[2], 'z_max_fixed_to_b']) for i in range(len(var_names))]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def evaluate(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        sigma_spin = params[self.hyper_var_names[1]]
        z_min = params[self.hyper_var_names[2]]
        prob  = truncnorm(data[self.var_names[0]], mu=1, sigma=sigma_spin, high=self.b, low=z_min)
        prob *= truncnorm(data[self.var_names[1]], mu=1, sigma=sigma_spin, high=self.b, low=z_min)
        prob *= xi_spin
        prob += (1-xi_spin)/((1-z_min)**2)
        return prob
    
    def __call__(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        sigma_spin = params[self.hyper_var_names[1]]
        z_min = params[self.hyper_var_names[2]]
        new_params = {**params, 'mu_spin':1, 'z_max_fixed_to_b' : self.b}
        prob  = self.models[0](data, new_params)           #truncnorm(data[self.var_names[0]], mu=1, sigma=sigma_spin, high=self.b, low=self.a)
        prob *= self.models[1](data, new_params)
        prob *= xi_spin
        prob += (1-xi_spin)/((1-z_min)**2)
        return prob


class TruncatedGaussian2DMixtureZeroAndFloatingIndependent(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['sigma_chi_at_0','mu_chi_1', 'sigma_chi_1', 'mu_chi_2', 'sigma_chi_2', 'rho_chi' 'eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.a = a; self.b=b;
        kwargs = {'a' : a, 'b' : b}
        comp1_var_names = [var_names[0]]
        comp2_var_names = [var_names[1]]

        comp1_hyper_var_names = ['mu_zero_spin_1d_fixed', self.hyper_var_names[0], 'mu_zero_spin_1d_fixed', self.hyper_var_names[0], 'rho_zero_fixed']
        comp2_hyper_var_names = self.hyper_var_names[1:6]
        self.mixture_hyper_var_name = self.hyper_var_names[-1]

        self.models = [TruncatedGaussian2DAnalytic(var_names=self.var_names, hyper_var_names=comp1_hyper_var_names, a=a, b=b),
                       TruncatedGaussian2DAnalytic(var_names=self.var_names, hyper_var_names=comp2_hyper_var_names, a=a, b=b)]

    @property
    def limits(self):
        return {var : [self.a[i], self.b[i]] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        params['rho_zero_fixed'] = 1e-6
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        params['rho_zero_fixed'] = 1e-6
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result


class TruncatedGaussian1DMixtureZeroAndFloatingIndependent(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['sigma_chi_at_0','mu_chi_1', 'sigma_chi_1', 'mu_chi_2', 'sigma_chi_2', 'eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.a = a; self.b=b;
        kwargs = {'a' : a, 'b' : b}
        comp1_var_names = [var_names[0]]
        comp2_var_names = [var_names[1]]

        comp1_hyper_var_names = ['mu_zero_spin_1d_fixed', self.hyper_var_names[0], 'mu_zero_spin_1d_fixed', self.hyper_var_names[0]]
        comp2_hyper_var_names = self.hyper_var_names[1:5]
        self.mixture_hyper_var_name = self.hyper_var_names[-1]

        self.models = [TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp1_hyper_var_names, a=a, b=b),
                       TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp2_hyper_var_names, a=a, b=b)]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result










####################################################################################################################
####  THE CODE BELOW IS TAKEN ALMOST ENTIRELY WORD FOR WORD FROM 
####  https://colab.research.google.com/drive/1w2tI1-1LWzPSdG_jE0FXzdrs6VAsJwhv?usp=sharing
####  It is an implementation by gihub user flaviovdf (https://github.com/flaviovdf)
####  of the paper "A simple approximation for the bivariate normal integral" by Wen-Jen Tsay & Peng-Hsuan Ke
####  https://doi.org/10.1080/03610918.2021.1884718
####################################################################################################################

import jax.numpy as jnp
import jax
from jax import jit
from jax.scipy.special import erf

from jax.scipy.stats.norm import cdf as cdf1d

@jax.jit
def case1(p, q, rho, a, b):
    c1 = -1.0950081470333
    c2 = -0.75651138383854
    sqrt2 = jnp.sqrt(2)
    a2 = a * a
    b2 = b * b

    line11 = 0.5 * (erf(q / sqrt2) + erf(b / (sqrt2 * a)))
    aux1 = a2 * c1 * c1 - 2 * sqrt2 * b * c1 + 2 * b2 * c2
    aux2 = 4 * (1 - a2 * c2)
    line12 = (1.0 / (4 * jnp.sqrt(1 - a2 * c2))) * jnp.exp(aux1 / aux2)

    line21 = 1 - erf((sqrt2 * b - a2 * c1) / (2 * a * jnp.sqrt(1 - a2 * c2)))
    aux3 = a2 * c1 * c1 + 2 * sqrt2 * b * c1 + 2 * b2 * c2
    line22 = (1.0 / (4 * jnp.sqrt(1 - a2 * c2))) * jnp.exp(aux3 / aux2)

    aux4 = sqrt2 * q - sqrt2 * a2 * c2 * q - sqrt2 * a * b * c2 - a * c1
    aux5 = 2 * jnp.sqrt(1 - a2 * c2)
    line31 = erf(aux4 / aux5)
    line32 = erf((a2 * c1 + sqrt2 * b) / (2 * a * jnp.sqrt(1 - a2 * c2)))

    return line11 + (line12 * line21) - (line22 * (line31 + line32))


@jax.jit
def case2(p, q):
    return cdf1d(p) * cdf1d(q)


@jax.jit
def case3(p, q, rho, a, b):
    c1 = -1.0950081470333
    c2 = -0.75651138383854
    sqrt2 = jnp.sqrt(2)
    a2 = a * a
    b2 = b * b

    aux1 = a2 * c1 * c1 - 2 * sqrt2 * b * c1 + 2 * b2 * c2
    aux2 = 4 * (1 - a2 * c2)
    line11 = (1.0 / (4 * jnp.sqrt(1 - a2 * c2))) * jnp.exp(aux1 / aux2)

    aux3 = sqrt2 * q - sqrt2 * a2 * c2 * q - sqrt2 * a * b * c2 + a * c1
    aux4 = 2 * jnp.sqrt(1 - a2 * c2)
    line12 = 1.0 + erf(aux3 / aux4)

    return line11 * line12


@jax.jit
def case4(p, q, rho, a, b):
    c1 = -1.0950081470333
    c2 = -0.75651138383854
    sqrt2 = jnp.sqrt(2)
    a2 = a * a
    b2 = b * b

    line11 = 0.5 + 0.5 * erf(q / sqrt2)
    aux1 = a2 * c1 * c1 + 2 * sqrt2 * b * c1 + 2 * b2 * c2
    aux2 = 4 * (1 - a2 * c2)
    line12 = (1.0 / (4 * jnp.sqrt(1 - a2 * c2))) * jnp.exp(aux1 / aux2)

    aux3 = sqrt2 * q - sqrt2 * a2 * c2 * q - sqrt2 * a * b * c2 - a * c1
    aux4 = 2 * jnp.sqrt(1 - a2 * c2)
    line2 = 1.0 + erf(aux3 / aux4)

    return line11 - (line12 * line2)


@jax.jit
def case5(p, q, rho, a, b):
    c1 = -1.0950081470333
    c2 = -0.75651138383854
    sqrt2 = jnp.sqrt(2)
    a2 = a * a
    b2 = b * b

    line11 = 0.5 - 0.5 * erf(b / (sqrt2 * a))
    aux1 = a2 * c1 * c1 + 2 * sqrt2 * b * c1 + 2 * b2 * c2
    aux2 = 4 * (1 - a2 * c2)
    line12 = (1.0 / (4 * jnp.sqrt(1 - a2 * c2))) * jnp.exp(aux1 / aux2)

    line21 = 1 - erf((sqrt2 * b + a2 * c1) / (2 * a * jnp.sqrt(1 - a2 * c2)))
    aux3 = a2 * c1 * c1 - 2 * sqrt2 * b * c1 + 2 * b2 * c2
    line22 = (1.0 / (4 * jnp.sqrt(1 - a2 * c2))) * jnp.exp(aux3 / aux2)

    aux4 = sqrt2 * q - sqrt2 * a2 * c2 * q - sqrt2 * a * b * c2 + a * c1
    aux5 = 2 * jnp.sqrt(1 - a2 * c2)
    line31 = erf(aux4 / aux5)
    line32 = erf((-a2 * c1 + sqrt2 * b) / (2 * a * jnp.sqrt(1 - a2 * c2)))

    return line11 - (line12 * line21) + line22 * (line31 + line32)


@jit
def binorm_lower(x1, x2, mu1, mu2, sigma1, sigma2, rho):
    p = (x1 - mu1) / sigma1
    q = (x2 - mu2) / sigma2

    a = -rho / jnp.sqrt(1 - rho * rho)
    b = p / jnp.sqrt(1 - rho * rho)
    
    cond1 = (a > 0) & (a * q + b >= 0)
    cond2 = (a == 0)
    cond3 = (a > 0) & (a * q + b < 0)
    cond4 = (a < 0) & (a * q + b >= 0)
    cond5 = (a < 0) & (a * q + b < 0)
    
    #print(cond1, cond2, cond3, cond4, cond5)
    
    return jnp.where(cond1, case1(p, q, rho, a, b),
             jnp.where(cond2, case2(p, q), 
                      jnp.where(cond3, case3(p, q, rho, a, b),
                               jnp.where(cond4, case4(p, q, rho, a, b),
                                        jnp.where(cond5, case5(p, q, rho, a, b), 
                                                 jnp.where(jnp.isposinf(p) & jnp.isposinf(q), jnp.ones_like(x1), 
                                                           jnp.zeros_like(x1)))))))

@jit
def binorm_upper(x1, x2, mu1, mu2, sigma1, sigma2, rho):
    p = -(x1 - mu1) / sigma1
    q = -(x2 - mu2) / sigma2

    a = -rho / jnp.sqrt(1 - rho * rho)
    b = p / jnp.sqrt(1 - rho * rho)
    
    cond1 = (a > 0) & (a * q + b >= 0)
    cond2 = (a == 0)
    cond3 = (a > 0) & (a * q + b < 0)
    cond4 = (a < 0) & (a * q + b >= 0)
    cond5 = (a < 0) & (a * q + b < 0)
    
    #print(cond1, cond2, cond3, cond4, cond5)
    
    return jnp.where(cond1, case1(p, q, rho, a, b),
             jnp.where(cond2, case2(p, q), 
                      jnp.where(cond3, case3(p, q, rho, a, b),
                               jnp.where(cond4, case4(p, q, rho, a, b),
                                        jnp.where(cond5, case5(p, q, rho, a, b), 
                                                 jnp.where(jnp.isneginf(p) & jnp.isneginf(q), jnp.ones_like(x1), 
                                                           jnp.zeros_like(x1)))))))


@jit
def mvnorm2d_using_lower(mu_1, mu_2, sigma_1, sigma_2, a_1, a_2, b_1, b_2, rho):
    #X1 = binorm_lower(b_1, b_2, mu_1, mu_2, sigma_1, sigma_2, rho)
    #X0 = binorm_upper(a_1, a_2, mu_1, mu_2, sigma_1, sigma_2, rho)
    #return X0-X1
    upper_right_lower = binorm_lower(b_1, b_2, mu_1, mu_2, sigma_1, sigma_2, rho)  ## X < b
    upper_left_lower  = binorm_lower(a_1, b_2, mu_1, mu_2, sigma_1, sigma_2, rho)  ## X < a_up
    lower_left_lower  = binorm_lower(a_1, a_2, mu_1, mu_2, sigma_1, sigma_2, rho)  ## X < a
    lower_right_lower = binorm_lower(b_1, a_2, mu_1, mu_2, sigma_1, sigma_2, rho)  ## X < b_down
    ## (b - b_down) + (a - a_up)
    return (upper_right_lower - lower_right_lower) + (lower_left_lower - upper_left_lower)

@jit
def mvnorm2d_using_upper(mu_1, mu_2, sigma_1, sigma_2, a_1, a_2, b_1, b_2, rho):
    #X1 = binorm_lower(b_1, b_2, mu_1, mu_2, sigma_1, sigma_2, rho)
    #X0 = binorm_upper(a_1, a_2, mu_1, mu_2, sigma_1, sigma_2, rho)
    #return X0-X1
    upper_right_upper = binorm_upper(b_1, b_2, mu_1, mu_2, sigma_1, sigma_2, rho)  ## X > b
    upper_left_upper  = binorm_upper(a_1, b_2, mu_1, mu_2, sigma_1, sigma_2, rho)  ## X > a_up
    lower_left_upper  = binorm_upper(a_1, a_2, mu_1, mu_2, sigma_1, sigma_2, rho)  ## X > a
    lower_right_upper = binorm_upper(b_1, a_2, mu_1, mu_2, sigma_1, sigma_2, rho)  ## X > b_down
    ## (b - b_down) + (a - a_up)
    return (upper_right_upper - lower_right_upper) + (lower_left_upper - upper_left_upper)


@jit
def mvnorm2d(mu_1, mu_2, sigma_1, sigma_2, a_1, a_2, b_1, b_2, rho):
    return jnp.where((mu_1 <= b_1) & (mu_2 <= b_2),  mvnorm2d_using_upper(mu_1, mu_2, sigma_1, sigma_2, a_1, a_2, b_1, b_2, rho), 
                        jnp.where((mu_1 >= a_1) & (mu_2 >= a_2), mvnorm2d_using_lower(mu_1, mu_2, sigma_1, sigma_2, a_1, a_2, b_1, b_2, rho),
                                 jnp.where((mu_1 <= a_1) & (mu_2 >= b_2), mvnorm2d_using_lower(b_1 + a_1 - mu_1, mu_2, sigma_1, sigma_2, a_1, a_2, b_1, b_2, -rho),
                                          mvnorm2d_using_upper(b_1 + a_1 - mu_1, mu_2, sigma_1, sigma_2, a_1, a_2, b_1, b_2, -rho))))



@jit
def mvnorm2d_deprecated(mu_1, mu_2, sigma_1, sigma_2, a_1, a_2, b_1, b_2, rho):
    #X1 = binorm_lower(b_1, b_2, mu_1, mu_2, sigma_1, sigma_2, rho)
    #X0 = binorm_upper(a_1, a_2, mu_1, mu_2, sigma_1, sigma_2, rho)
    #return X0-X1
    upper_right_lower = binorm_lower(b_1, b_2, mu_1, mu_2, sigma_1, sigma_2, rho)
    upper_left_lower  = binorm_lower(a_1, b_2, mu_1, mu_2, sigma_1, sigma_2, rho)
    lower_left_lower  = binorm_lower(a_1, a_2, mu_1, mu_2, sigma_1, sigma_2, rho)
    lower_right_lower = binorm_lower(b_1, a_2, mu_1, mu_2, sigma_1, sigma_2, rho)
    return (upper_right_lower - lower_right_lower) - (upper_left_lower - lower_left_lower)





import h5py
import jax.numpy as jnp

def load_hdf5_to_jax_dict(filename, ignore_events=[]):
    data = {}
    with h5py.File(filename, 'r') as f:
        for event_name, event_group in f.items():
            event_data = {}
            if event_name in ignore_events:
                continue
            for dataset_name, dataset in event_group.items():
                if isinstance(dataset, h5py.Dataset):
                    event_data[dataset_name] = jnp.array(dataset)
                elif isinstance(dataset, h5py.Group):
                    subgroup_data = {}
                    for subgroup_name, subgroup_dataset in dataset.items():
                        subgroup_data[subgroup_name] = jnp.array(subgroup_dataset)
                    event_data[dataset_name] = subgroup_data
            data[event_name] = event_data
    return data

def load_hdf5_attributes(filename):
    data = {}
    with h5py.File(filename, 'r') as f:
        for group_name, group in f.items():
            group_data = {}
            for attr_name, attr_value in group.attrs.items():
                # Check the datatype of the attribute value
                if isinstance(attr_value, bytes):
                    # If it's bytes, decode it to string
                    group_data[attr_name] = attr_value.decode("utf-8")
                else:
                    # Otherwise, store it as is
                    group_data[attr_name] = attr_value
            data[group_name] = group_data
    return data


def stack_nested_jax_arrays(data):
    stacked_data = {}
    eventnames = []
    for event_name, event_data in data.items():
        eventnames.append(event_name)
        for var_name, var_data in event_data.items():
            if isinstance(var_data, dict):
                if var_name not in stacked_data:
                    stacked_data[var_name] = stack_nested_jax_arrays({event_name: var_data})[0]
                else:
                    stacked_data[var_name].update(stack_nested_jax_arrays({event_name: var_data})[0])
            else:
                if var_name not in stacked_data:
                    stacked_data[var_name] = jnp.expand_dims(var_data, axis=0)
                else:
                    stacked_data[var_name] = jnp.vstack((stacked_data[var_name], jnp.expand_dims(var_data, axis=0)))
    return stacked_data, eventnames


import jax
import jax.numpy as jnp
from ..utils import *
from ..generic import *



def alpha_beta_max_to_mu_var_max(alpha, beta, amax):
    mu = alpha / (alpha + beta) * amax
    var = alpha * beta / ((alpha + beta) ** 2 * (alpha + beta + 1)) * amax ** 2
    return mu, var, amax


def mu_var_max_to_alpha_beta_max(mu, var, amax):
    mu /= amax
    var /= amax ** 2
    alpha = (mu ** 2 * (1 - mu) - mu * var) / var
    beta = (mu * (1 - mu) ** 2 - (1 - mu) * var) / var
    return alpha, beta, amax


class GaussianIsotropicSpinOrientationsIID(SampledPopulationModel, SpinPopulationModel):
    r"""
    Mixture of gaussian and isotropic distribution over spin orientations.
    Performs a monte carlo estimate of the population likelihood. 

    Event Level Parameters:         :math:`z_1=cos(\theta_1), z_2=cos(\theta_2)`
    Population Level Parameters:    :math:`\xi, \sigma` 

    .. math::
    
        P(z_1, z_2| \xi, \sigma) = \frac{1-\xi}{4} + \xi \left(  \mathcal{N}_{[-1,1]}(z_1 | \xi, \sigma) \mathcal{N}_{[-1,1]}(z_2 | \xi, \sigma) \right) 
    """
    def __init__(self, var_names=['cos_tilt_1', 'cos_tilt_2'], hyper_var_names=['xi_spin','sigma_spin'], a=-1, b=1):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    
    def __call__(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        sigma_spin = params[self.hyper_var_names[1]]
        prob  = truncnorm(data[self.var_names[0]], mu=1, sigma=sigma_spin, high=self.b, low=self.a)
        prob *= truncnorm(data[self.var_names[1]], mu=1, sigma=sigma_spin, high=self.b, low=self.a)
        prob *= xi_spin
        prob += (1-xi_spin)/4
        return prob


class GaussianIsotropicSpinOrientationsIIDWithMinimum(SampledPopulationModel, SpinPopulationModel):
    r"""
    Mixture of gaussian and isotropic distribution over spin orientations.
    Performs a monte carlo estimate of the population likelihood. 

    Event Level Parameters:         :math:`z_1=cos(\theta_1), z_2=cos(\theta_2)`
    Population Level Parameters:    :math:`\xi, \sigma` 

    .. math::
    
        P(z_1, z_2| \xi, \sigma) = \frac{1-\xi}{4} + \xi \left(  \mathcal{N}_{[-1,1]}(z_1 | \xi, \sigma) \mathcal{N}_{[-1,1]}(z_2 | \xi, \sigma) \right) 
    """
    def __init__(self, var_names=['cos_tilt_1', 'cos_tilt_2'], hyper_var_names=['xi_spin','sigma_spin', 'z_min'], a=-1, b=1):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    
    def __call__(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        sigma_spin = params[self.hyper_var_names[1]]
        z_min = params[self.hyper_var_names[2]]
        prob  = truncnorm(data[self.var_names[0]], mu=1, sigma=sigma_spin, high=self.b, low=self.a)
        prob *= truncnorm(data[self.var_names[1]], mu=1, sigma=sigma_spin, high=self.b, low=self.a)
        prob *= xi_spin
        prob += jnp.heaviside(data[self.var_names[0]] - z_min, 0.5)*jnp.heaviside(data[self.var_names[1]] - z_min, 0.5)*(1-xi_spin)/4
        return prob

class GaussianIsotropicSpinOrientationsIIDExtended(SampledPopulationModel, SpinPopulationModel):
    r"""
    Mixture of gaussian and isotropic distribution over spin orientations.
    Performs a monte carlo estimate of the population likelihood. 

    Event Level Parameters:         :math:`z_1=cos(\theta_1), z_2=cos(\theta_2)`
    Population Level Parameters:    :math:`\xi, \sigma` 

    .. math::
    
        P(z_1, z_2| \xi, \sigma) = \frac{1-\xi}{4} + \xi \left(  \mathcal{N}_{[-1,1]}(z_1 | \xi, \sigma) \mathcal{N}_{[-1,1]}(z_2 | \xi, \sigma) \right) 
    """
    def __init__(self, var_names=['cos_tilt_1', 'cos_tilt_2'], hyper_var_names=['xi_spin','sigma_spin_1', 'sigma_spin_2', 'z_min_1', 'z_min_2'], a=-1, b=1):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        sigma_spin_1 = params[self.hyper_var_names[1]]
        sigma_spin_2 = params[self.hyper_var_names[2]]
        z_min_1 = params[self.hyper_var_names[3]]
        z_min_2 = params[self.hyper_var_names[4]]
        prob  = truncnorm(data[self.var_names[0]], mu=1, sigma=sigma_spin_1, high=self.b, low=z_min_1)
        prob *= truncnorm(data[self.var_names[1]], mu=1, sigma=sigma_spin_2, high=self.b, low=z_min_2)
        prob *= xi_spin
        prob += ((1-xi_spin)) * jnp.heaviside(data[self.var_names[0]] - z_min_1, 0.5) * jnp.heaviside(data[self.var_names[1]] - z_min_2, 0.5) / ((1-z_min_1)*(1-z_min_2))
        return prob


class GaussianIsotropicSpinOrientationsIIDFullExtended(SampledPopulationModel, SpinPopulationModel):
    r"""
    Mixture of gaussian and isotropic distribution over spin orientations.
    Performs a monte carlo estimate of the population likelihood. 

    Event Level Parameters:         :math:`z_1=cos(\theta_1), z_2=cos(\theta_2)`
    Population Level Parameters:    :math:`\xi, \sigma` 

    .. math::
    
        P(z_1, z_2| \xi, \sigma) = \frac{1-\xi}{4} + \xi \left(  \mathcal{N}_{[-1,1]}(z_1 | \xi, \sigma) \mathcal{N}_{[-1,1]}(z_2 | \xi, \sigma) \right) 
    """
    def __init__(self, var_names=['cos_tilt_1', 'cos_tilt_2'], hyper_var_names=['xi_spin','sigma_spin', 'z_min'], a=-1, b=1):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    
    def __call__(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        sigma_spin = params[self.hyper_var_names[1]]
        z_min = params[self.hyper_var_names[2]]
        prob1  = xi_spin*(1/(1-z_min))*jnp.where(data[self.var_names[0]] < z_min, 0, 1) + (1-xi_spin)*truncnorm(data[self.var_names[0]], mu=1, sigma=sigma_spin, high=self.b, low=z_min)
        prob2  = xi_spin*(1/(1-z_min))*jnp.where(data[self.var_names[1]] < z_min, 0, 1) + (1-xi_spin)*truncnorm(data[self.var_names[1]], mu=1, sigma=sigma_spin, high=self.b, low=z_min)
        return prob1*prob2


class GaussianIsotropicSpinOrientationsFloating(SampledPopulationModel, SpinPopulationModel):
    r"""
    Mixture of gaussian and isotropic distribution over spin orientations.
    Performs a monte carlo estimate of the population likelihood. 

    Event Level Parameters:         :math:`z_1=cos(\theta_1), z_2=cos(\theta_2)`
    Population Level Parameters:    :math:`\xi, \sigma` 

    .. math::
    
        P(z_1, z_2| \xi, \sigma) = \frac{1-\xi}{4} + \xi \left(  \mathcal{N}_{[-1,1]}(z_1 | \xi, \sigma) \mathcal{N}_{[-1,1]}(z_2 | \xi, \sigma) \right) 
    """
    def __init__(self, var_names=['cos_tilt_1', 'cos_tilt_2'], hyper_var_names=['xi_spin','mu_spin_1', 'sigma_spin_1', 'mu_spin_2', 'sigma_spin_2'], a=-1, b=1):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    
    def __call__(self, data, params):
        xi_spin = params[self.hyper_var_names[0]]
        mu_spin_1, mu_spin_2 = params[self.hyper_var_names[1]], params[self.hyper_var_names[3]]
        sigma_spin_1, sigma_spin_2 = params[self.hyper_var_names[2]], params[self.hyper_var_names[4]]
        prob  = truncnorm(data[self.var_names[0]], mu=mu_spin_1, sigma=sigma_spin_1, high=self.b, low=self.a)
        prob *= truncnorm(data[self.var_names[1]], mu=mu_spin_2, sigma=sigma_spin_2, high=self.b, low=self.a)
        prob *= xi_spin
        prob += (1-xi_spin)/4
        return prob



def sample_beta(alpha, beta, amax, size=1):
    """
    Generate samples from a scaled beta distribution.

    Parameters:
    alpha (np.array): Alpha parameter of the Beta distribution.
    beta (np.array): Beta parameter of the Beta distribution.
    amax (np.array): Maximum value to scale the Beta distribution to.
    size (int): Number of samples to generate (default is 1).

    Returns:
    np.array: Scaled samples from the Beta distribution.
    """
    # Ensure alpha, beta, and amax are numpy arrays
    alpha = np.asarray(alpha)
    beta = np.asarray(beta)
    amax = np.asarray(amax)

    # Generate beta samples (values between 0 and 1)
    samples = np.random.beta(alpha, beta, size=size)
    
    # Scale samples by amax
    scaled_samples = samples * amax

    return scaled_samples


class BetaSpinMagnitudeIID(SampledPopulationModel, SpinPopulationModel):
    r"""
    Mixture of gaussian and isotropic distribution over spin orientations.
    Performs a monte carlo estimate of the population likelihood. 

    Event Level Parameters:         :math:`z_1=cos(\theta_1), z_2=cos(\theta_2)`
    Population Level Parameters:    :math:`\xi, \sigma` 

    .. math::
    
        P(z_1, z_2| \xi, \sigma) = \frac{1-\xi}{4} + \xi \left(  \mathcal{N}_{[-1,1]}(z_1 | \xi, \sigma) \mathcal{N}_{[-1,1]}(z_2 | \xi, \sigma) \right) 
    """
    def __init__(self, var_names=['a_1', 'a_2'], hyper_var_names=['mu_chi','sigma_chi','amax'],parameterization="mu_sigma", constraints=None):
        self.parameterization = parameterization
        self.converter = lambda x,y,z: (x,y,z)
        self.var_names = var_names
        self.constraints = constraints
        if self.constraints is not None:
            self.alpha_min_constraint = constraints['alpha'][0]
            self.alpha_max_constraint = constraints['alpha'][1]
            self.beta_min_constraint = constraints['beta'][0]
            self.beta_max_constraint = constraints['beta'][1]
        self.hyper_var_names = hyper_var_names
        if (len(self.hyper_var_names) == 2) and 'amax' not in self.hyper_var_names:
            self.hyper_var_names = self.hyper_var_names + ['amax']

        ### SIGMA_CHI IS ACTUALLY SIGMA_CHI**2 THIS IS UNFORTUNATE

        hyper_var_names_are_mu_sigma = (("mu" in self.hyper_var_names[0]) or (("sigma" in self.hyper_var_names[1])))

        if (parameterization in ("mu_sigma")) and hyper_var_names_are_mu_sigma:
            self.converter = lambda mu,sigma,amax=1 : mu_var_max_to_alpha_beta_max(mu, sigma, amax)

    @property
    def limits(self):
        return {var : [0, 1] for i,var in enumerate(self.var_names)}
    
    def __call__(self, data, params):
        amax = params.get(self.hyper_var_names[2], 1)
        alpha_chi, beta_chi, amax = self.converter(params[self.hyper_var_names[0]], params[self.hyper_var_names[1]], amax)
        if self.constraints is not None:
            is_alpha_within_constraint = (alpha_chi > self.alpha_min_constraint) & (alpha_chi < self.alpha_max_constraint)
            is_beta_within_constraint = (beta_chi > self.beta_min_constraint) & (beta_chi < self.beta_max_constraint)
            are_both_within_constraint = is_alpha_within_constraint & is_beta_within_constraint
        prob  = beta(data[self.var_names[0]], alpha=alpha_chi, beta=beta_chi, scale=amax)
        prob *= beta(data[self.var_names[1]], alpha=alpha_chi, beta=beta_chi, scale=amax)
        if self.constraints is not None:
            prob = jnp.where(are_both_within_constraint, prob, jnp.nan_to_num(-jnp.inf)*jnp.ones_like(prob))
        return prob

    def sample(self, df_hyper_samples, oversample=1):
        amax = df_hyper_samples.get(self.hyper_var_names[2], 1)
        alpha_chi, beta_chi, amax = self.converter(df_hyper_samples[self.hyper_var_names[0]], df_hyper_samples[self.hyper_var_names[1]], amax)
        chi_1 = sample_beta(alpha_chi, beta_chi, amax, size=oversample);
        chi_2 = sample_beta(alpha_chi, beta_chi, amax, size=oversample);
        return pd.DataFrame({self.var_names[0] : chi_1, self.var_names[1] : chi_2})




import jax.numpy as jnp
import jax
from jax import jit

#def box(x,a,b):
#    return jnp.heaviside(x-a, 0.5)*jnp.heaviside(b-x, 0.5)

@jit
def box(x,a,b):
    return jnp.where(x >= a, jnp.where(x <= b, jnp.ones_like(x), jnp.zeros_like(x)), jnp.zeros_like(x))

@jit
def powerlaw(x, alpha, high, low):
    norm = jnp.where(
        jnp.array(alpha) == -1,
        1 / jnp.log(high / low),
        (1 + alpha) / jnp.array(high ** (1 + alpha) - low ** (1 + alpha)),
    )
    prob = jnp.power(x, alpha)*norm*box(x, low, high)
    return prob

@jit
def truncnorm(x, mu, sigma, high, low):
    norm = 2**0.5 / jnp.pi**0.5 / sigma
    norm /= jax.scipy.special.erf((high - mu) / 2**0.5 / sigma) + jax.scipy.special.erf(
        (mu - low) / 2**0.5 / sigma
    )
    prob = jnp.exp(-jnp.power(x - mu, 2) / (2 * sigma**2))
    prob *= norm*box(x, low, high)
    return prob


import jax.numpy as jnp
from jax import grad
from jax.scipy.special import betaln

@jit
def beta(x, alpha, beta, scale):
    ln_beta = (alpha - 1) * jnp.log(x) + (beta - 1) * jnp.log(scale - x)
    ln_beta -= betaln(alpha, beta)
    ln_beta -= (alpha + beta - 1) * jnp.log(scale)
    prob = jnp.exp(ln_beta)
    prob *= box(x, 0.0, scale)
    prob = jnp.nan_to_num(prob)
    return prob



from gravpop import *
import gravpop
from gwpopulation.models.redshift import PowerLawRedshift
from bilby.core.result import read_in_result
from scipy.interpolate import interp1d
import numpy as np
import deepdish as dd
import matplotlib.pyplot as plt
import pandas as pd
from tqdm.auto import trange
from matplotlib.gridspec import GridSpec

PP_path = '/Users/asadh/Documents/Data/analyses/PowerLawPeak/o1o2o3_mass_c_iid_mag_iid_tilt_powerlaw_redshift_result.json'
PP_result = read_in_result(PP_path)

PP_hyperposterior_samples = PP_result.posterior.copy().sample(100) # making a copy is best practice here so you don't accidentally modify things in-place
PP_hyperposterior_samples = BETA_CONVERTER.convert_parameters(PP_hyperposterior_samples, remove=False)


class TestAgainstGWPop:
	def setup_method(self, method):
		models = dict(
						mass="SmoothedMassDistribution",#"two_component_primary_mass_ratio",
						mag="iid_spin_magnitude",
						tilt="iid_spin_orientation",
						redshift="gwpopulation.models.redshift.PowerLawRedshift",
					)
		gwpop = GWPopLoader(posterior_file = "/Users/asadh/Documents/Data/posteriors.pkl",
							prior_file = "/Users/asadh/Documents/Data/production.prior",
							vt_file = "/Users/asadh/Downloads/o1+o2+o3_bbhpop_real+semianalytic-LIGO-T2100377-v2.hdf5",
							enforce_minimum_neffective_per_event = "injection_resampling_vt",
							models = models, vt_models=models,
							max_redshift = 2.3, samples_per_posterior=5000)
		gwpop.load_prior();
		gwpop.load_model();
		gwpop.load_vt();
		gwpop.load_posteriors();
		gwpop.create_likelihood();
		self.gwpop = gwpop

		SM = SmoothedTwoComponentPrimaryMassRatio(primary_mass_name='mass_1')
		R = gravpop.PowerLawRedshift(z_max=2.3)
		S_mag = BetaSpinMagnitudeIID(var_names = ['a_1', 'a_2'])
		S_tilt = GaussianIsotropicSpinOrientationsIID(var_names = ['cos_tilt_1', 'cos_tilt_2'])


		event_data = gwpop.likelihood.data.copy()
		event_data['prior'] = gwpop.likelihood.sampling_prior
		selection_data = {k:v for k,v in gwpop.selection.data.items()}

		self.gravpop_likelihood = PopulationLikelihood(models=[SM,R,S_mag,S_tilt], 
													   event_data=event_data,
													   selection_data=selection_data,
													   analysis_time=selection_data['analysis_time'],
													   total_generated=selection_data['total_generated'])

		self.HP = HyperPosterior(posterior = PP_hyperposterior_samples,
								 likelihood = self.gravpop_likelihood,
								 mass_model = SM,
								 redshift_model = R,
								 rate = False)

		self.HP.posterior = self.HP.posterior.rename({'min_event_n_effective' : 'min_event_n_effective_lvk',
													  'pdet_n_effective' : 'pdet_n_effective_gwpop_lvk'}, axis=1)

		self.HP.calculate_rates()
		self.HP.N_effective_cuts()

	def get_model_like(self, model_list, target_name):
		names = [obj.__name__ if hasattr(obj, '__name__') else obj.__class__.__name__ for obj in model_list]
		return [model for name, model in zip(names, model_list) if target_name in name.lower()][0]

	def evaluate_gwpop_mass_model_at(self, param):
		#model = self.gwpop.models['mass']
		#model = [x for x in self.gwpop.likelihood.hyper_prior.models if "mass" in x.__class__.__name__.lower()][0]
		model = self.get_model_like(self.gwpop.likelihood.hyper_prior.models, "mass")
		self.gwpop.likelihood.hyper_prior.parameters.update(param)
		params = self.gwpop.likelihood.hyper_prior._get_function_parameters(model)
		return model(self.gwpop.likelihood.data, **params)

	def evaluate_gravpop_mass_model_at(self, param):
		model = self.gravpop_likelihood.models[0]
		return model(self.gravpop_likelihood.event_data, param)

	def evaluate_gwpop_redshift_model_at(self, param):
		#model = self.gwpop.models['redshift']
		#model = [x for x in self.gwpop.likelihood.hyper_prior.models if "redshift" in x.__class__.__name__.lower()][0]
		model = self.get_model_like(self.gwpop.likelihood.hyper_prior.models, "redshift")
		self.gwpop.likelihood.hyper_prior.parameters.update(param)
		params = self.gwpop.likelihood.hyper_prior._get_function_parameters(model)
		return model(self.gwpop.likelihood.data, **params)

	def evaluate_gravpop_spin_orientation_model_at(self, param):
		model = self.gravpop_likelihood.models[3]
		return model(self.gravpop_likelihood.event_data, param)

	def evaluate_gwpop_spin_orientation_model_at(self, param):
		#model = self.gwpop.models['redshift']
		#model = [x for x in self.gwpop.likelihood.hyper_prior.models if "spin_orientation" in x.__class__.__name__.lower()][0]
		model = self.get_model_like(self.gwpop.likelihood.hyper_prior.models, "spin_orientation")
		self.gwpop.likelihood.hyper_prior.parameters.update(param)
		params = self.gwpop.likelihood.hyper_prior._get_function_parameters(model)
		return model(self.gwpop.likelihood.data, **params)

	def evaluate_gravpop_spin_magnitude_model_at(self, param):
		model = self.gravpop_likelihood.models[2]
		return model(self.gravpop_likelihood.event_data, param)

	def evaluate_gwpop_spin_magnitude_model_at(self, param):
		#model = self.gwpop.models['redshift']
		#model = [x for x in self.gwpop.likelihood.hyper_prior.models if "spin_magnitude" in x.__class__.__name__.lower()][0]
		model = self.get_model_like(self.gwpop.likelihood.hyper_prior.models, "spin_magnitude")
		self.gwpop.likelihood.hyper_prior.parameters.update(param)
		params = self.gwpop.likelihood.hyper_prior._get_function_parameters(model)
		return model(self.gwpop.likelihood.data, **params)

	def evaluate_gravpop_redshift_model_at(self, param):
		model = self.gravpop_likelihood.models[1]
		return model(self.gravpop_likelihood.event_data, param)

	def test_compare_redshift(self):
		cols = ['alpha','beta','mmax','mmin','lam','mpp','sigpp','delta_m','lamb']
		param = PP_hyperposterior_samples[cols].iloc[0,:].to_dict()

		Z_gwpop = jnp.nan_to_num(self.evaluate_gwpop_redshift_model_at(param), posinf=0, neginf=0, nan=0)
		Z_gravpop = jnp.nan_to_num(self.evaluate_gravpop_redshift_model_at(param), posinf=0, neginf=0, nan=0)

		rel = jnp.abs((Z_gravpop))/(jnp.abs(Z_gwpop) + 1e-16)

		fig = plt.figure(layout="constrained")
		gs = GridSpec(1, 1, figure=fig)
		ax3 = fig.add_subplot(gs[0, 0])
		ax3.hist(rel.flatten(), bins=100);
		ax3.set_title(r"Histogram of $P_{gravpop}/P_{gwpop}$")
		plt.suptitle("Probability comparison for redshift magnitude models")
		fig.savefig("./test/images/data_redshift_magnitude_model_comparison.png")
		assert not jnp.any(jnp.isnan(Z_gravpop)) # TESTING NO NANs


	def test_compare_redshift(self):
		cols = ['alpha','beta','mmax','mmin','lam','mpp','sigpp','delta_m','lamb']
		param = PP_hyperposterior_samples[cols].iloc[0,:].to_dict()

		Z_gwpop = jnp.nan_to_num(self.evaluate_gwpop_redshift_model_at(param), posinf=0, neginf=0, nan=0)
		Z_gravpop = jnp.nan_to_num(self.evaluate_gravpop_redshift_model_at(param), posinf=0, neginf=0, nan=0)

		rel = jnp.abs((Z_gravpop))/(jnp.abs(Z_gwpop) + 1e-16)

		fig = plt.figure(layout="constrained")
		gs = GridSpec(1, 1, figure=fig)
		ax3 = fig.add_subplot(gs[0, 0])
		ax3.hist(rel.flatten(), bins=100);
		ax3.set_title(r"Histogram of $P_{gravpop}/P_{gwpop}$")
		plt.suptitle("Probability comparison for redshift magnitude models")
		fig.savefig("./test/images/data_redshift_magnitude_model_comparison.png")
		assert not jnp.any(jnp.isnan(Z_gravpop)) # TESTING NO NANs

	def test_compare_spin_orientation(self):
		cols = ['alpha','beta','mmax','mmin','lam','mpp','sigpp','delta_m','lamb',
		'mu_chi', 'sigma_chi', 'xi_spin', 'sigma_spin', 'lamb', 'amax','alpha_chi', 'beta_chi']
		param = PP_hyperposterior_samples[cols].iloc[0,:].to_dict()

		Z_gwpop = jnp.nan_to_num(self.evaluate_gwpop_spin_orientation_model_at(param), posinf = 0, neginf = 0, nan=0)
		Z_gravpop = jnp.nan_to_num(self.evaluate_gravpop_spin_orientation_model_at(param), posinf = 0, neginf = 0, nan=0)

		#rel = jnp.abs((Z_gwpop - Z_gravpop))/(jnp.abs(Z_gwpop) + 1e-16)

		rel = jnp.abs((Z_gravpop))/(jnp.abs(Z_gwpop) + 1e-16)

		fig = plt.figure(layout="constrained")
		gs = GridSpec(1, 1, figure=fig)
		ax3 = fig.add_subplot(gs[0, 0])
		ax3.hist(rel.flatten(), bins=100);
		ax3.set_title(r"Histogram of $P_{gravpop}/P_{gwpop}$")
		plt.suptitle("Probability comparison for spin orientation models")
		fig.savefig("./test/images/data_spin_orientation_model_comparison.png")
		#assert not jnp.any(jnp.isnan(Z_gravpop)) # TESTING NO NANs
		#assert rel.max() < 1e-2

	def test_compare_spin_magnitude(self):
		cols = ['alpha','beta','mmax','mmin','lam','mpp','sigpp','delta_m','lamb',
		'mu_chi', 'sigma_chi', 'xi_spin', 'sigma_spin', 'lamb', 'amax','alpha_chi', 'beta_chi']
		param = PP_hyperposterior_samples[cols].iloc[0,:].to_dict()

		Z_gwpop = jnp.nan_to_num(self.evaluate_gwpop_spin_magnitude_model_at(param), posinf = 0, neginf = 0, nan=0)
		Z_gravpop = jnp.nan_to_num(self.evaluate_gravpop_spin_magnitude_model_at(param), posinf = 0, neginf = 0, nan=0)

		#rel = jnp.abs((Z_gwpop - Z_gravpop))/(jnp.abs(Z_gwpop) + 1e-16)

		rel = jnp.abs((Z_gravpop))/(jnp.abs(Z_gwpop) + 1e-16)

		fig = plt.figure(layout="constrained")
		gs = GridSpec(1, 1, figure=fig)
		ax3 = fig.add_subplot(gs[0, 0])
		ax3.hist(rel.flatten(), bins=100);
		ax3.set_title(r"Histogram of $P_{gravpop}/P_{gwpop}$")
		plt.suptitle("Probability comparison for spin magnitude models")
		fig.savefig("./test/images/data_spin_magnitude_model_comparison.png")
		#assert not jnp.any(jnp.isnan(Z_gravpop)) # TESTING NO NANs
		#assert rel.max() < 1e-2

	def test_event_N_effective(self):
		cols = ['alpha', 'beta', 'mmax', 'mmin', 'lam', 'mpp', 'sigpp', 'delta_m',
				'mu_chi', 'sigma_chi', 'xi_spin', 'sigma_spin', 'lamb', 'amax','alpha_chi', 'beta_chi']

		param = self.HP.posterior[cols].iloc[0,:].to_dict()

		self.gwpop.likelihood.parameters.update(param)

		col_new = 'min_event_n_effective_gwpop'
		new_data = {col_new : np.zeros_like(self.HP.posterior['min_event_n_effective'].values)}
		for i in trange(len(self.HP.posterior)):
			param = self.HP.posterior[cols].iloc[i,:].to_dict()
			self.gwpop.likelihood.parameters.update(param)
			new_data[col_new][i] = jnp.min(self.gwpop.likelihood.per_event_bayes_factors_and_n_effective_and_variances()[1])
			
			
		self.HP.posterior.loc[:, col_new] = new_data[col_new];

		delta_N_eff = self.HP.posterior['min_event_n_effective_gwpop'] - self.HP.posterior['min_event_n_effective']

		fig = plt.figure(layout="constrained")
		gs = GridSpec(1, 1, figure=fig)
		ax3 = fig.add_subplot(gs[0, 0])
		ax3.hist(delta_N_eff.values, bins=10);
		ax3.set_title(r"Histogram of $N^{gravpop}_{eff}/N^{gwpop}_{eff}$")
		plt.suptitle(r"$N_{eff}$ comparison")
		fig.savefig("./test/images/data_event_N_eff_magnitude_model_comparison.png")

		#assert (().abs() < 5).all()

	def test_selection_N_effective(self):
		cols = ['alpha', 'beta', 'mmax', 'mmin', 'lam', 'mpp', 'sigpp', 'delta_m',
				'mu_chi', 'sigma_chi', 'xi_spin', 'sigma_spin', 'lamb', 'amax','alpha_chi', 'beta_chi']

		param = self.HP.posterior[cols].iloc[0,:].to_dict()

		self.gwpop.likelihood.parameters.update(param)

		col_selection_new = 'pdet_n_effective_gwpop'
		new_data = {col_selection_new : np.zeros_like(self.HP.posterior['pdet_n_effective'].values)}
		for i in trange(len(self.HP.posterior)):
			param = self.HP.posterior[cols].iloc[i,:].to_dict()
			self.gwpop.likelihood.parameters.update(param)
			mu, var = self.gwpop.selection.detection_efficiency(param)
			N_eff_sel = mu**2 / var
			new_data[col_selection_new][i] = N_eff_sel

		self.HP.posterior.loc[:, col_selection_new] = new_data[col_selection_new];
		
		delta_N_eff = self.HP.posterior['pdet_n_effective_gwpop'] - self.HP.posterior['pdet_n_effective']

		#assert ((self.HP.posterior['min_event_n_effective_gwpop'] - self.HP.posterior['min_event_n_effective']).abs() < 5).all()
		#assert ((self.HP.posterior['pdet_n_effective_gwpop'] - self.HP.posterior['pdet_n_effective']).abs() < 5).all()
		
		fig = plt.figure(layout="constrained")
		gs = GridSpec(1, 1, figure=fig)
		ax3 = fig.add_subplot(gs[0, 0])
		ax3.hist(delta_N_eff.values, bins=10);
		ax3.set_title(r"Histogram of Selection $N^{gwpop}_{eff} - N^{gravpop}_{eff}$")
		plt.suptitle(r"Selection $N_{eff}$ comparison")
		fig.savefig("./test/images/data_selection_N_eff_magnitude_model_comparison.png")

	def teardown_method(self, method):
		pass

	def test_mass_model_matches_gwpop(self):
		pass



import jax
import jax.numpy as jnp
from .abstract import *
from ..utils import *

class Gaussian1D(SampledPopulationModel):
    r"""
    Gaussian Distribution. Performs a monte carlo estimate of the population likelihood. 

    Event Level Parameters:         :math:`x`
    Population Level Parameters:    :math:`\mu, \sigma` 

    .. math::
    
        P(x | \mu, \sigma) = \mathcal{N}_{[a,b]}))(x | \mu, \sigma) 
    """
    def __init__(self, var_name='x', mu_name='mu', sigma_name='sigma'):
        self.var_name = var_name
        self.mu_name = mu_name
        self.sigma_name = sigma_name
        
    def get_data(self, data, params):
        Xs          = data[self.var_name];
        mu          = params[self.mu_name]
        sigma       = params[self.sigma_name]
        return Xs, mu, sigma
    
    def __call__(self, data, params):
        Xs, mu, sigma = self.get_data(data, params);
        pdf = jax.scipy.stats.norm.pdf(Xs, loc=mu, scale=sigma).mean(axis=-1)
        return pdf


@jit
def transform_gaussians_gauss(x,dx,mu,sigma):
    #x = K.x; Δx = K.Δx;
    denom = dx**2 + sigma**2
    x_ = (dx**2 * mu + sigma**2 * x)/denom;
    sigma_ = jnp.sqrt((sigma**2 * dx**2)/denom);
    return x_, sigma_

@jit
def logweight_gauss(x,dx,mu,sigma,a,b):
    x_, sigma_ = transform_gaussians(x,dx,mu,sigma);
    
    logC_ = logC(x_, sigma_, a, b); #print(logC_)
    logCmu = logC(mu ,sigma , a, b);# print(logCmu)
    #logCx = logC(x ,dx , a, b); #print(logCx)
    logA = logC_ - logCmu# - logCx
    return logA

@jit
def loglikelihood_kernel1d_gauss(x, dx, mu, sigma, a, b):
    logA = logweight_gauss(x,dx,mu,sigma,a,b)
    logB = jax.scipy.stats.norm.logpdf(x,loc=mu, scale=jnp.sqrt(dx**2 + sigma**2))     #normlogpdf(x,jnp.sqrt(Δx^2 + σ^2), μ)
    return logA + logB


class Gaussian1DAnalytic1D(AnalyticPopulationModel):
    r"""
    Gaussian Distribution. Performs a monte carlo estimate of the population likelihood. 

    Event Level Parameters:         :math:`x`
    Population Level Parameters:    :math:`\mu, \sigma` 

    .. math::
    
        P(x | \mu, \sigma) = \mathcal{N}_{[a,b]}))(x | \mu, \sigma) 
    """
    def __init__(self, var_name='x', mu_name='mu', sigma_name='sigma'):
        self.var_name = var_name
        self.mu_name = mu_name
        self.sigma_name = sigma_name
        
    def get_data(self, data, params):
        Xs          = data[self.var_name];
        mu          = params[self.mu_name]
        sigma       = params[self.sigma_name]
        return Xs, mu, sigma
    
    def __call__(self, data, params):
        Xs, mu, sigma = self.get_data(data, params);
        pdf = jax.scipy.stats.norm.pdf(Xs, loc=mu, scale=sigma).mean(axis=-1)
        return pdf



class Gaussian1DAnalytic(AnalyticPopulationModel):
    def __init__(self, a, b, var_names=['x'], hyper_var_names=['mu', 'sigma']):
        self.a = a
        self.b = b
        self.var_name = var_names[0]
        self.mu_name = hyper_var_names[0]
        self.sigma_name = hyper_var_names[1]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}
        
    def get_data(self, data, params):
        X_locations = data[self.var_name + '_mu_kernel'];
        X_scales    = data[self.var_name + '_sigma_kernel'];
        mu          = params[self.mu_name]
        sigma       = params[self.sigma_name]
        return X_locations, X_scales, mu, sigma

    def evaluate(self, data, params):
        """
        Evaluates the value of the *integrand* and not the integral
        """
        Xs          = data[self.var_name];
        mu          = params[self.mu_name]
        sigma       = params[self.sigma_name];
        #loglikes = jax.scipy.special.logsumexp( jnp.log(truncnorm(Xs, mu, sigma, low=a, high=b)) , axis=-1) - jnp.log(Xs.shape[-1])
        prob = truncnorm(Xs, mu, sigma, low=self.a, high=self.b)
        return prob
        
    def __call__(self, data, params):
        X_locations, X_scales, mu, sigma = self.get_data(data, params);
        loglikes = loglikelihood_kernel1d_gauss(X_locations, X_scales, mu, sigma, self.a, self.b)
        return jnp.exp(loglikes)

    def sample(self, df_hyper_samples, oversample=1):
        return ppd_truncCorrelatedanalytic(self, df_hyper_samples, oversample=oversample)




class Gaussian1DAnalytic(AnalyticPopulationModel):
    def __init__(self, a, b, var_names=['x'], hyper_var_names=['mu', 'sigma']):
        self.a = a
        self.b = b
        self.var_name = var_names[0]
        self.mu_name = hyper_var_names[0]
        self.sigma_name = hyper_var_names[1]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}
        
    def get_data(self, data, params):
        X_locations = data[self.var_name + '_mu_kernel'];
        X_scales    = data[self.var_name + '_sigma_kernel'];
        mu          = params[self.mu_name]
        sigma       = params[self.sigma_name]
        return X_locations, X_scales, mu, sigma

    def evaluate(self, data, params):
        """
        Evaluates the value of the *integrand* and not the integral
        """
        Xs          = data[self.var_name];
        mu          = params[self.mu_name]
        sigma       = params[self.sigma_name];
        #loglikes = jax.scipy.special.logsumexp( jnp.log(truncnorm(Xs, mu, sigma, low=a, high=b)) , axis=-1) - jnp.log(Xs.shape[-1])
        prob = truncnorm(Xs, mu, sigma, low=self.a, high=self.b)
        return prob
        
    def __call__(self, data, params):
        X_locations, X_scales, mu, sigma = self.get_data(data, params);
        loglikes = loglikelihood_kernel1d_gauss(X_locations, X_scales, mu, sigma, self.a, self.b)
        return jnp.exp(loglikes)

    def sample(self, df_hyper_samples, oversample=1):
        return ppd_truncCorrelatedanalytic(self, df_hyper_samples, oversample=oversample)


import jax
import jax.numpy as jnp
from jax import jit
from ..models.utils import *
from ..models.redshift import Redshift
from ..utils import *
from .selection import SelectionFunction
import numpy as np


from dataclasses import dataclass, field
from typing import List, Dict, Optional, Union
from functools import partial


@dataclass
class PopulationLikelihood:
    models:   List  # List of population models evaluated using monte-carlo
    event_data:       Dict[str, jax.Array] = field(repr=False)    # Dictionary of data e.g. {'mass_1' : jnp.array([21.2, 23.0, ...], ....)}
    selection_data:   Optional[Union[Dict[str, jax.Array], SelectionFunction]] = field(default=None, repr=False)   # Dictionary of data e.g. {'mass_1' : jnp.array([21.2, 23.0, ...], ....)}
    analysis_time: float = 1 # Analysis duration, default to one year
    total_generated: Optional[int] = None
    enforce_convergence: bool = False
    event_names : Optional[List[str]] = None
    
    def __post_init__(self):
        #print("redoing prior")
        if "prior" not in self.event_data:
            print("Warning: Prior not specified... setting prior to 1")
            keyvalues = list(self.event_data.items())
            largest_shape = keyvalues[np.argmax([len(value.shape) for key,value in keyvalues])][1].shape # most probably a sampled variable
            self.event_data["prior"] = jnp.ones(shape=largest_shape)
        self.N_events = self.event_data['prior'].shape[0]
        self.log_epsilon = 1e-30
        self.redshift_model = [model for model in self.models if isinstance(model, Redshift)]
        if len(self.redshift_model) == 0:
            self.redshift_model = None
        else:
            self.redshift_model = self.redshift_model[0]

        if isinstance(self.selection_data, Dict):
            self.selection_data = SelectionFunction(self.selection_data, self.analysis_time, self.total_generated, self.redshift_model)
        elif isinstance(self.selection_data, SelectionFunction):
            self.analysis_time = self.selection_data.analysis_time
            self.total_generated = self.selection_data.total_generated

        self.N_posterior_samples = self.event_data["prior"].shape[-1]
        if len(self.models) == 0:
            raise ValueError("We've got no models to work with :(")
        #print("finished making likleihood")
    
    @staticmethod
    def log(x):
        return jnp.log(x + 1e-37)

    def sampled_compute_log_weights(self, data, params):
        return sum(self.log(model(data, params)) for model in self.models) - self.log(data["prior"]) # E x N 
    
    def sampled_event_bayes_factors(self, data, params, N=None):
        logweights = self.sampled_compute_log_weights(data, params)
        N = N or logweights.shape[-1]
        loglikes = jax.scipy.special.logsumexp(logweights, axis=-1) - jnp.log(N)  # E
        return loglikes

    def total_event_bayes_factors(self, data, params, N=None, selection=False):
        if selection and self.enforce_convergence:
            logweights = self.sampled_compute_log_weights(data, params)
            log_mu, N_eff = self._compute_selection_N_eff(logweights)
            return jnp.where(N_eff > 4*self.N_events, log_mu, jnp.nan_to_num(-jnp.inf))
        elif (not selection) and self.enforce_convergence:
            logweights = self.sampled_compute_log_weights(data, params)
            log_mu, N_eff = self._compute_event_min_N_eff(logweights, N=N)
            return jnp.where(N_eff > self.N_events, log_mu, jnp.nan_to_num(-jnp.inf))
        else:
            return self.sampled_event_bayes_factors(data, params, N=N)

    def _compute_selection_N_eff(self, logweights, N=None):
        N = N or self.selection_data.total_generated
        log_mu = jax.scipy.special.logsumexp(logweights, axis=-1) - jnp.log(N)
        mu = jnp.exp(log_mu)
        log_var_1 = jax.scipy.special.logsumexp(2*logweights , axis=-1) - 2*jnp.log(N)
        log_var_2 = 2*log_mu - jnp.log(N)
        var = jnp.exp(log_var_1) - jnp.exp(log_var_2)
        N_eff = mu**2 / var
        return log_mu, N_eff

    def _compute_event_min_N_eff(self, logweights, N=None):
        N = N or logweights.shape[-1]
        log_mu = jax.scipy.special.logsumexp(logweights, axis=-1)
        log_N_eff = 2*log_mu - jax.scipy.special.logsumexp(2*logweights , axis=-1)
        N_eff = jnp.min(jnp.exp(log_N_eff), axis=-1)
        return log_mu- jnp.log(N), N_eff

    def compute_selection_N_eff(self, params, N=None):
        logweights = self.sampled_compute_log_weights(self.selection_data.selection_data, params)
        return self._compute_selection_N_eff(logweights, N=N)

    def compute_event_N_eff(self, params, N=None):
        N = N or self.N_posterior_samples
        logweights = self.sampled_compute_log_weights(self.event_data, params)
        return self._compute_event_min_N_eff(logweights, N=N)

    def compute_selection_N_eff_only(self, params, N=None):
        logweights = self.sampled_compute_log_weights(self.selection_data.selection_data, params)
        return self._compute_selection_N_eff(logweights, N=N)[1]

    def compute_event_N_eff_only(self, params, N=None):
        N = N or self.N_posterior_samples
        logweights = self.sampled_compute_log_weights(self.event_data, params)
        return self._compute_event_min_N_eff(logweights, N=N)[1]
    
    def logpdf(self, params):
        # Event Likelihoods
        loglikes_event =self.total_event_bayes_factors(self.event_data, params, N=self.N_posterior_samples)
        
        # Selection Likelihood
        loglikes_selection = 0.0
        if self.selection_data:
            loglikes_selection += self.total_event_bayes_factors(self.selection_data.selection_data, params, 
                                                                 N=self.selection_data.total_generated,
                                                                 selection=True)
        
        return jnp.nan_to_num( loglikes_event.sum() - self.N_events * loglikes_selection )

    @classmethod
    def from_file(cls, event_data_filename, selection_data_filename, models, SelectionClass=SelectionFunction, enforce_convergence=False, samples_per_event=1000):
        event_file_ext = event_data_filename.split('.')[-1]
        if event_file_ext in ('hdf5', 'h5'):
            # Extract the entire hdf5 as a nested dictionary of jax arrays
            event_data, event_names = stack_nested_jax_arrays(load_hdf5_to_jax_dict(event_data_filename))
        elif event_file_ext == 'pkl':
            # Extract a posteriors.pkl file usually an intermediate data product of gwpopulation_pipe
            # Must be a pickle of a list of pandas dataframes
            import numpy as np
            import pandas as pd
            df_list = np.load(event_data_filename, allow_pickle=True)
            minimum_length = min([min(len(df),samples_per_event) for df in df_list])
            event_data = {col : jnp.stack([df_list[i][col][0:minimum_length].values for i in range(len(df_list))]) for col in df_list[0].columns}
            event_data['mass_1_source'] = event_data['mass_1']
            event_data['chi_1'] = event_data['a_1']
            event_data['chi_2'] = event_data['a_2']
            event_names = list(range(len(event_data)))
        print(list(event_data.keys()))
        print("(N_events, N_samples_per_event) = ", event_data['prior'].shape)
        selection_data = load_hdf5_to_jax_dict(selection_data_filename)
        selection_attributes = load_hdf5_attributes(selection_data_filename)

        #print("reordering data...")
        if "selection" in selection_data.keys():
            selection_data = selection_data["selection"]

        if "selection" in selection_attributes.keys():
            selection_attributes = selection_attributes["selection"]

        redshift_model = [model for model in models if isinstance(model, Redshift)]
        if len(redshift_model) == 0:
            redshift_model = None
        else:
            redshift_model = redshift_model[0]

        #print("creating selection")
        selection = SelectionClass(selection_data, 
                                   selection_attributes['analysis_time'],
                                   selection_attributes['total_generated'],
                                   redshift_model)
        return cls(models, event_data, selection, enforce_convergence=enforce_convergence, event_names=event_names)



### Test Data
from dataclasses import dataclass
import numpyro.distributions as dist
from numpyro.distributions import Distribution
import jax
import jax.numpy as jnp
from jax import jacfwd, jacrev
from gravpop.utils.vmap import chunked_vmap
import numpy as np

from typing import Dict, List

def jax_to_numpy(obj):
    if isinstance(obj, jnp.ndarray):
        return np.array(obj)
    elif isinstance(obj, dict):
        return {key: jax_to_numpy(value) for key, value in obj.items()}
    elif isinstance(obj, (list, tuple)):
        return type(obj)(jax_to_numpy(item) for item in obj)
    else:
        return obj

def model_gradient(model, data, param, canonical_parameter_order=None):
	canonical_parameter_order = canonical_parameter_order or list(param.keys())

	def make_vector(d):
		return jnp.array([d[param] for param in canonical_parameter_order])

	def make_dictionary(x):
		return {parameter : x[i] for i,parameter in enumerate(canonical_parameter_order)}

	#dYdx = jacrev(lambda x: model(data, make_dictionary(x)))(make_vector(param))
	dYdx = jacfwd(lambda x: model(data, make_dictionary(x)))(make_vector(param))
	if len(canonical_parameter_order) == 1:
		return {parameter : dYdx.flatten() for i,parameter in enumerate(canonical_parameter_order)}

	return {parameter : dYdx[..., i] for i,parameter in enumerate(canonical_parameter_order)}

@dataclass
class PriorDictionary:
	priors : Dict[str, Distribution]
	seed : int = 0

	def sample(seed=None):
		seed = seed or self.seed
		key = jax.random.key(seed)
		point = {}
		for variable, prior_distribution in self.priors.items():
			point[variable] = prior_distribution.sample(key)

		return point


class Dirac(dist.Distribution):
    arg_constraints = {}
    support = dist.constraints.real
    has_enumerate_support = False

    def __init__(self, value):
        super(Dirac, self).__init__(event_shape=(1,))
        self.value = value

    def sample(self, key, sample_shape=()):
        return jnp.full(sample_shape + self.event_shape, self.value)

    def log_prob(self, value):
        return jnp.where(value == self.value, 0., -jnp.inf)

@dataclass
class Grid1D:
	name : str = 'x'
	minimum : float = 0.0
	maximum : float = 0.0
	N : int = 100
	latex_name : str = r'$x$'

	def __post_init__(self):
		if self.name != 'x':
			self.latex_name = fr'${self.name}$'

		self.grid = jnp.linspace(self.minimum, self.maximum, self.N)
		self._data = {self.name : self.grid}

	@property
	def data(self):
		return self._data
	

@dataclass
class Grid:
	grid_list : List[Grid1D]

	def __post_init__(self):
		self.grid_1ds = {g.name : g.grid for g in self.grid_list}
		grid_combos = jnp.meshgrid(*[g.grid for g in self.grid_list])
		data = {}
		for i,g in enumerate(self.grid_list):
			data[g.name] = grid_combos[i]
		self._data = data

	@property
	def data(self):
		return self._data


GLOBAL_DICTIONARY_OF_LIMITS = {
    'alpha' : [-4,12],
    'beta' : [-2,7],
    'mmin' : [2,10],
    'mmax' : [30,100],
    'delta_m' : [0,10],
    'lam' : [0,1],
    'mpp' : [20,50],
    'sigpp' : [1,10],
    'lamb' : [-6,6],
    'mu_1' : [0,1],
    'sigma_1' : [0,5],
    'mu_2' : [0,1],
    'sigma_2' : [0,5]
}



GLOBAL_DICTIONARY_OF_NICE_VALUES = {
    'alpha' : 3.5,
    'beta' : 1.1,
    'mmin' : 5,
    'mmax' : 95,
    'delta_m': 0.1,
    'lam' : 0.4,
    'mpp' : 35,
    'sigpp' : 4,
    'lamb' : 2.9,
    'mu_1' : 0.1,
    'sigma_1' : 0.1,
    'mu_2' : 0.1,
    'sigma_2' : 0.1
}


GLOBAL_DICTIONARY_OF_EVIL_VALUES = {
	 'alpha': 3.5,
	 'beta': 1.1,
	 'delta_m': 3.0,
	 'lam': 0.4,
	 'mmin': 3.0,
	 'mmax': 95.0,
	 'mpp': 35.0,
	 'sigpp': 4.0,
	 'lamb': 2.9
}


PRIOR_DICT = PriorDictionary({var : dist.Uniform(*limits) for var, limits in GLOBAL_DICTIONARY_OF_LIMITS.items()})




#{'alpha':3.5, 'beta':1.1, 'delta_m':0.03, 'mpp': 30.0, 'sigpp': 3.0, 'lam':0.04, 'mmin':2.0, 'mmax': 90.0}
		





import jax
import jax.numpy as jnp

def convert_log_weights_to_weights(logweight):
	if logweight is not None:
		return jnp.exp(logweight)
	else:
		return  None


class AbstractExpectation:
	"""
	Abstract class to compute the expectation of the likelihood integral
	in different cases
	"""
	def __init__(self):
		pass

	def num_samples(self, weights, N_samples=None):
		return N_samples or weights.shape[-1]

####################################################
### Sampled Expectation
###################################################

class SampledExpectation(AbstractExpectation): # takes in an array with shape E x ... x N
	def log_mu(self, logweights, N_samples=None):
		if N_samples is None:
			N_samples = self.num_samples(logweights, N_samples)
		return jax.scipy.special.logsumexp(logweights, axis=-1) - jnp.log(N_samples)  # E x ...

	def mu(self, logweights, N_samples):
		return jnp.exp(logweights).sum(axis=-1) / N_samples  

	def log_bayes_factors(self, logweights, N_samples=None):
		return self.log_mu(logweights, N_samples=N_samples)


class SampledEventExpectation(SampledExpectation):
	def var(self, logweights, N_samples=None):
		weights = jnp.exp(logweights)
		N = self.num_samples(weights, N_samples=N_samples)
		var = (weights**2).sum(axis=-1)
		var /= N**2
		return var # E

	def N_eff(self, logweights, mu=None, N_samples=None):
		weights = jnp.exp(logweights)
		N_samples = self.num_samples(weights, N_samples=N_samples)
		if mu is None:
			mu = self.mu(weights, N_samples=N_samples)
		var = self.var(weights, N_samples=N_samples)
		return mu**2 / var

class SampledSelectionExpectation(SampledExpectation):
	def var(self, logweights, mu=None, N_samples=None):
		weights = jnp.exp(logweights)
		if mu is None:
			mu = self.mu(weights, N_samples=None)
		N = self.num_samples(weights, N_samples=N_samples)
		var = (weights**2).sum(axis=-1)
		var /= N**2
		var -= (mu**2 / N)
		return var

	def N_eff(self, logweights, mu=None, N_samples=None):
		weights = jnp.exp(logweights)
		N_samples = self.num_samples(weights, N_samples=N_samples)
		if mu is None:
			mu = (weights.sum(axis=-1) / N_samples)
		var = self.var(weights, mu=mu, N_samples=N_samples)
		return mu**2 / var

####################################################
### Analytic Expectation
###################################################

class AnalyticExpectation(AbstractExpectation): # E x ... x K
	def num_components(self, logweights, N_components=None):
		return N_components or logweights.shape[-1]

	def log_mu(self, logweights):
		return logweights					    # E x ... x K

	def mu(self, logweights):
		return jnp.exp(log_weights)

	def aggregate_components(self, loglikes, weights=1):
		return jax.scipy.special.logsumexp( loglikes + jnp.log(weights), axis=-1) # E x ...

	def log_bayes_factors(self, logweights, weights=1):
		return self.aggregate_components(self.log_mu(logweights), weights) # E x ...

####################################################
### Hybrid Expectation
###################################################

class HybridExpectation(SampledExpectation, AnalyticExpectation): # E x ... x K x N
	def num_components(self, sampled_logweights=None, analytic_logweights=None, weights=None, N_components=None):
		if N_components is not None:
			return N_components
		if weights is not None:
			return weights.shape[-1]
		if analytic_logweights is not None:
			return analytic_logweights.shape[-1]
		if sampled_logweights is not None:
			return sampled_logweights.shape[-2]
		return N_components

	def log_mu(self, sampled_logweights=None, analytic_logweights=None, N_samples_per_component=None):
		log_bayes_factor_per_component = 0.0
		if sampled_logweights is not None:
			N_samples_per_component = self.num_samples(sampled_logweights, N_samples=N_samples_per_component)
			#print(SampledExpectation.log_mu(self, sampled_logweights, N_samples=N_samples_per_component).shape)
			log_bayes_factor_per_component += SampledExpectation.log_mu(self, sampled_logweights, N_samples=N_samples_per_component)
		if analytic_logweights is not None:
			#print(analytic_logweights.shape)
			#print(AnalyticExpectation.log_mu(self, analytic_logweights).shape)
			log_bayes_factor_per_component += AnalyticExpectation.log_mu(self, analytic_logweights)
		return log_bayes_factor_per_component

	def mu(self, sampled_logweights=None, analytic_logweights=None, N_samples_per_component=None):
		return jnp.exp(self.log_mu(sampled_logweights, analytic_logweights, N_samples_per_component))

	def log_bayes_factors(self, sampled_logweights=None, analytic_logweights=None, weights=1, N_samples_per_component=None):
		log_bayes_factor_per_component = self.log_mu(sampled_logweights, analytic_logweights, N_samples_per_component)
		log_bayes_factors = AnalyticExpectation.aggregate_components(self, log_bayes_factor_per_component, weights=weights)
		return log_bayes_factors


class HybridEventExpectation(HybridExpectation):
	def var(self, sampled_logweights=None, analytic_logweights=None, N_samples_per_component=None, weights=1):
		sampled_weights = jnp.exp(sampled_logweights) if sampled_logweights is not None else None
		analytic_weights = jnp.exp(analytic_logweights) if analytic_logweights is not None else None
		N_k = self.num_samples(sampled_weights, N_samples=N_samples_per_component)
		K = self.num_components(sampled_logweights, analytic_logweights, weights)
		var = (sampled_weights**2).sum(axis=-1)  # E x ... x K x N -> E x ... x K
		var *= (analytic_weights * weights)**2   # E x ... x K
		var = var.sum(axis=-1) / N_k**2 # E x ... 
		return var

	def N_eff(self, sampled_logweights=None, analytic_logweights=None, mu=None, N_samples_per_component=None, weights=1):
		N_samples_per_component = self.num_samples(sampled_logweights, N_samples=N_samples_per_component)
		if mu is None:
			mu = jnp.exp(self.log_bayes_factors(sampled_logweights=sampled_logweights, 
										   analytic_logweights=analytic_logweights, 
										   N_samples_per_component=N_samples_per_component,
										   weights=weights))
		var = self.var(sampled_logweights=sampled_logweights, 
					   analytic_logweights=analytic_logweights, 
					   N_samples_per_component=N_samples_per_component, 
					   weights=weights)
		return mu**2 / var


class HybridSelectionExpectation(HybridExpectation):
	def log_mu(self, sampled_logweights=None, analytic_logweights=None, N_samples_per_component=None, total_generated=None):
		if total_generated is not None:
			K = self.num_components(sampled_logweights, analytic_logweights)
			N_k = self.num_samples(sampled_logweights, N_samples=N_samples_per_component)
			detection_ratio = N_k * K / total_generated
		else:
			detection_ratio = 1
		return HybridExpectation.log_mu(self, sampled_logweights=sampled_logweights, analytic_logweights=analytic_logweights, 
								N_samples_per_component=N_samples_per_component) + jnp.log(detection_ratio)

	def log_bayes_factors(self, sampled_logweights=None, analytic_logweights=None, weights=1, N_samples_per_component=None, total_generated=None):
		log_bayes_factor_per_component = self.log_mu(sampled_logweights, analytic_logweights, N_samples_per_component)
		log_bayes_factors = AnalyticExpectation.aggregate_components(self, log_bayes_factor_per_component, weights=weights)
		return log_bayes_factors

	def mu(self, sampled_logweights=None, analytic_logweights=None, N_samples_per_component=None, total_generated=None):
		return jnp.exp(self.log_mu(sampled_logweights=sampled_logweights, analytic_logweights=analytic_logweights, 
								   N_samples_per_component=N_samples_per_component, total_generated=total_generated))

	def var(self, mu=None, sampled_logweights=None, analytic_logweights=None, N_samples_per_component=None, weights=1, total_generated=None):
		if mu is None:
			mu = self.log_bayes_factors(sampled_logweights=sampled_logweights, 
									   analytic_logweights=analytic_logweights, 
									   N_samples_per_component=N_samples_per_component,
									   weights=weights, total_generated=total_generated)
		sampled_weights = jnp.exp(sampled_logweights) if sampled_logweights is not None else None
		analytic_weights = jnp.exp(analytic_logweights) if analytic_logweights is not None else None
		K = self.num_components(sampled_logweights, analytic_logweights, weights)
		N_k = self.num_samples(sampled_weights, N_samples=N_samples_per_component)
		detection_ratio = N_k * K / total_generated
		var = (sampled_weights**2).sum(axis=-1)  # E x ... x K x N -> E x ... x K
		var *= (analytic_weights * weights)**2   # E x ... x K
		var = detection_ratio**2 * var.sum(axis=-1) / N_k # E x ... 
		var -= (mu**2 / total_generated)  # E x ... 
		return var

	def N_eff(self, sampled_logweights=None, analytic_logweights=None, mu=None, N_samples_per_component=None, weights=1, total_generated=None):
		N_samples_per_component = self.num_samples(sampled_logweights, N_samples=N_samples_per_component)
		if mu is None:
			mu = jnp.exp(self.log_bayes_factors(sampled_logweights=sampled_logweights, 
										   analytic_logweights=analytic_logweights, 
										   N_samples_per_component=N_samples_per_component,
										   weights=weights, total_generated=total_generated))
		var = self.var(sampled_logweights=sampled_logweights, 
					   analytic_logweights=analytic_logweights, 
					   mu=mu, N_samples_per_component=N_samples_per_component, 
					   weights=weights, total_generated=total_generated)
		return mu**2 / var





from .abstract import *
import numpy as np
import jax.numpy as jnp
import pandas as pd

def make_list_if_not(x):
    if not isinstance(x, list):
        return [x]

class ProductModel(AbstractPopulationModel):
    def __init__(self, models):
        self.var_names = []
        self.hyper_var_names = []
        self.a = []; self.b = [];
        for model in models:
            self.var_names = self.var_names + model.var_names
            self.hyper_var_names = self.hyper_var_names + model.hyper_var_names
            self.a = self.a + make_list_if_not(model.a);
            self.b = self.b + make_list_if_not(model.b);

        self.models = models

    @property
    def limits(self):
        lims = {}
        for model in self.models:
            lims = {**lims, **model.limits}
        return lims

    def __call__(self, data, params):
        result  = self.models[0](data, params);
        for i in range(1,len(self.models)):
            result *= self.models[i](data, params);
        return result

    def evaluate(self, data, params):
        result  = self.models[0].evaluate(data, params);
        for i in range(1,len(self.models)):
            result *= self.models[i].evaluate(data, params);
        return result

    def sample(self, df_hyper_samples, oversample=1, **kwargs):
        kwargs['oversample'] = oversample
        series_s = []
        for model in self.models:
            series_s.append(model.sample(df_hyper_samples, **kwargs).reset_index(drop=True))

        N = len(df_hyper_samples)
        return pd.concat(series_s, axis=1)

class Mixture2D(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, model1, model2, mixture_hyper_var_name="eta_spin", var_names=None, hyper_var_names=None):
        if var_names is not None:
            self.var_names = model1.var_names
        else:
            self.var_names = model1.var_names
        self.hyper_var_names = model1.hyper_var_names + model2.hyper_var_names + [mixture_hyper_var_name]
        self.a = model1.a; self.b = model1.b;

        self.mixture_hyper_var_name = mixture_hyper_var_name

        self.models = [model1, model2]

    @property
    def limits(self):
        return self.models[0].limits

    def __call__(self, data, params):
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples, oversample=1, **kwargs):
        kwargs['oversample'] = oversample
        series_1 = self.models[0].sample(df_hyper_samples, **kwargs).reset_index(drop=True)
        series_2 = self.models[1].sample(df_hyper_samples, **kwargs).reset_index(drop=True)
        N = len(df_hyper_samples)
        sampled_mixture = np.hstack([(df_hyper_samples[self.mixture_hyper_var_name] >= np.random.rand(N)) for _ in range(oversample)])
        d = {col : ((sampled_mixture * series_1[col]) + ((1 - sampled_mixture) * series_2[col])) for col in series_1.columns}
        return pd.DataFrame(d)

class SampledFromAnalytic(SampledPopulationModel):
    def __init__(self, analytic_model):
        self.analytic_model = analytic_model
        self.__dict__.update(analytic_model.__dict__)
        
    def __call__(self, data, params):
        return self.analytic_model.evaluate(data, params)
        
    def evaluate(self, data, params):
        return self.analytic_model.evaluate(data, params)

class FixedParameters(AnalyticPopulationModel):
    def __init__(self, model, fixed_parameters : dict):
        self.model = model
        self.fixed_parameters = fixed_parameters
        for attr, value in model.__dict__.items():
            setattr(self, attr, value)

    @property
    def limits(self):
        return self.model.limits

    def evaluate(self, data, params):
        return self.model.evaluate(data, {**params, **self.fixed_parameters})

    def __call__(self, data, params):
        return self.model(data, {**params, **self.fixed_parameters})

    def sample(self, df_hyper_samples, **kwargs):
        df2 = df_hyper_samples.copy()
        for p,v in self.fixed_parameters.items():
            df2[p] = v
        return self.model.sample(df2, **kwargs)


# Spin Tests
from gravpop import *
import gwpopulation
import matplotlib.pyplot as plt
import matplotlib
from matplotlib.gridspec import GridSpec
import numpy as np

## Test Non Analytic Spin that it integrates to one
def test_non_analytic_spin_integrates_to_one():
    TG = TruncatedGaussian1D(a=0, b=1, var_name='x')
    params = {'mu' : 0.0, 'sigma': 0.1}
    test = lambda x: TG({'x' : jnp.array([x])}, params)

    xs = np.linspace(0,1,10_000)
    ys = np.array([test(x) for x in xs])
    #plt.plot(xs, ys)

    assert np.isclose( np.trapz(y=ys, x=xs), 1.0, rtol=1e-1)


## Test Analytic Spin matches expected result from monte carlo estimate
def test_analytic_spin_matches_monte_carlo_estimate():
    import scipy
    TG = TruncatedGaussian1D(a=0, b=1)
    TGA = TruncatedGaussian1DAnalytic(a=0, b=1)
    mu_0, sigma_0 = 0.5, 0.1
    data = {'x' : scipy.stats.truncnorm.rvs(loc=0.5, scale=0.1, a=(0-mu_0)/sigma_0, b=(1-mu_0)/sigma_0, size=800_000)}
    data['x_mu_kernel'] = mu_0
    data['x_sigma_kernel'] = sigma_0


    params = {'mu' : 0.5, 'sigma': 0.1}
    assert jnp.isclose( TG(data, params), TGA(data, params) , rtol=1e-3)


class TestSampledSpinModels:

    def setup_method(self, method):
        self.evaluation_point = {'mu_chi': 2.0, 'sigma_chi':0.9, 'xi_spin':0.1, 'sigma_spin':0.8}

    def create_magnitude_dataset(self):
        a1_grid = Grid1D("a_1", minimum=0, maximum=1, N=100)
        a2_grid = Grid1D("a_2", minimum=0, maximum=1, N=100)
        a_grid = Grid([a1_grid, a2_grid])
        return a_grid.data

    def create_orientations_dataset(self):
        cos_tilt_1_grid = Grid1D("cos_tilt_1", minimum=-1, maximum=1, N=100)
        cos_tilt_2_grid = Grid1D("cos_tilt_2", minimum=-1, maximum=1, N=100)
        cos_tilt_grid = Grid([cos_tilt_1_grid, cos_tilt_2_grid])
        return cos_tilt_grid.data

    def gwpopulation_iid_magnitude_model(self, data, params):
        cols = ['alpha_chi', 'beta_chi']
        return gwpopulation.models.spin.iid_spin_magnitude_beta(data, **{k:v for k,v in params.items() if k in cols})

    def gwpopulation_iid_orientations_model(self, data, params):
        cols = ['xi_spin', 'sigma_spin']
        return gwpopulation.models.spin.iid_spin_orientation_gaussian_isotropic(data, **{k:v for k,v in params.items() if k in cols})

    def gravpop_iid_magnitude_model(self, data, params):
        Mag = BetaSpinMagnitudeIID(parameterization="mu_sigma", var_names = ['a_1', 'a_2'])
        return Mag(data, params)

    def gravpop_iid_orientations_model(self, data, params):
        Orient = GaussianIsotropicSpinOrientationsIID(var_names = ['cos_tilt_1', 'cos_tilt_2'])
        return Orient(data, params)

    def test_compare_magnitudes(self):
        mag_data = self.create_magnitude_dataset()
        Z_gwpop = self.gwpopulation_iid_magnitude_model(mag_data, self.evaluation_point)
        Z_gravpop = self.gravpop_iid_magnitude_model(mag_data, self.evaluation_point)

        fig = plt.figure(layout="constrained")
        gs = GridSpec(2, 2, figure=fig)
        ax1 = fig.add_subplot(gs[0, 0])
        ax2 = fig.add_subplot(gs[0, 1])
        ax3 = fig.add_subplot(gs[1, :])
        ax1.contourf(mag_data['a_1'], mag_data['a_2'], jnp.log(Z_gravpop))
        ax2.contourf(mag_data['a_1'], mag_data['a_2'], jnp.log(Z_gwpop))
        ax1.set_xlabel(r"$a_1$"); ax2.set_xlabel(r"$a_1$");
        ax1.set_ylabel(r"$a_2$"); ax2.set_ylabel(r"$a_2$");
        ax1.set_title(r"$\log P_{gravpop}$")
        ax2.set_title(r"$\log P_{gwpopulation}$")
        ax3.hist(jnp.abs(Z_gravpop/(jnp.abs(Z_gwpop) + 1e-16)).flatten(), bins=10);
        ax3.set_title(r"Histogram of $P_{gravpop}/P_{gwpop}$")
        plt.suptitle("Log probability comparison for spin magnitude models")
        fig.savefig("./test/images/spin_magnitude_model_comparison.png")
        assert not jnp.any(jnp.isnan(Z_gravpop)) # TESTING NO NANs

    def test_compare_orientations(self):
        orientation_data = self.create_orientations_dataset()
        Z_gwpop = self.gwpopulation_iid_orientations_model(orientation_data, self.evaluation_point)
        Z_gravpop = self.gravpop_iid_orientations_model(orientation_data, self.evaluation_point)

        fig = plt.figure(layout="constrained")
        gs = GridSpec(2, 2, figure=fig)
        ax1 = fig.add_subplot(gs[0, 0])
        ax2 = fig.add_subplot(gs[0, 1])
        ax3 = fig.add_subplot(gs[1, :])
        ax1.contourf(orientation_data['cos_tilt_1'], orientation_data['cos_tilt_2'], jnp.log(Z_gravpop))
        ax2.contourf(orientation_data['cos_tilt_1'], orientation_data['cos_tilt_2'], jnp.log(Z_gwpop))
        ax1.set_xlabel(r"$cos(\theta_1)$"); ax2.set_xlabel(r"$cos(\theta_1)$");
        ax1.set_ylabel(r"$cos(\theta_2)$"); ax2.set_ylabel(r"$cos(\theta_2)$");
        ax1.set_title(r"$\log P_{gravpop}$")
        ax2.set_title(r"$\log P_{gwpopulation}$")
        ax3.hist(jnp.abs(Z_gravpop/(jnp.abs(Z_gwpop) + 1e-16)).flatten() - 1, bins=10);
        ax3.set_title(r"Histogram of $P_{gravpop}/P_{gwpop} - 1$")
        plt.suptitle("Log probability comparison for spin orientation models")
        fig.savefig("./test/images/spin_orientations_model_comparison.png")
        assert not jnp.any(jnp.isnan(Z_gravpop)) # TESTING NO NANs









from typing import Union, Dict, List, Optional
from dataclasses import dataclass, field
import pandas as pd
from .grid import Grid1D, Grid
from ..utils.vmap import chunked_vmap
from ..models import AbstractPopulationModel
from ..hyper import PopulationLikelihood
import jax
import jax.numpy as jnp
import numpy as np

DEFAULT_GRID =  Grid([Grid1D(name='mass_1_source', minimum=2, maximum=100, N=200, latex_name=r"$m_1$"), 
                      Grid1D(name='mass_ratio', minimum=0, maximum=1  , N=200, latex_name=r"$q$")])


@dataclass
class MassPlot:
    hyper_posterior_samples : Dict[str, jax.Array] = field(repr=False)
    model : Optional[AbstractPopulationModel] = None
    mass_grid : Optional[Dict[str, Union[Grid, Grid1D]]] = None
    confidence_interval : float = 0.95
    rate : bool = False
    chunk : int = 20
    n_samples : int = 1000
    
    def __post_init__(self):
        acceptable_sample_names = self.model.hyper_var_names + ['rate']
        total_indices = self.hyper_posterior_samples[next(iter(self.hyper_posterior_samples.keys()))].size
        self.index_sampling = np.random.randint(total_indices, size=self.n_samples)
        self.hyper_posterior_samples = {col:value[self.index_sampling] for col,value in self.hyper_posterior_samples.items() if col in acceptable_sample_names}
        self._shapes = {key:0 for key in self.hyper_posterior_samples.keys()}
        self.mass_grid = self.mass_grid or DEFAULT_GRID 
        data = self.mass_grid.data
        self.result = None
        progress_title = "Computing Mass Model on the Grid"
        self._vmapped_func = chunked_vmap( lambda x: self.model.evaluate(data, x), in_axes=(self._shapes,), chunk=self.chunk, progress_note=progress_title)
        self.conf = ((1-self.confidence_interval)/2, self.confidence_interval + (1-self.confidence_interval)/2)

    def compute(self):
        self.result = self._vmapped_func(self.hyper_posterior_samples)
        if self.rate:
            self.result = self.result * self.hyper_posterior_samples["rate"][..., None, None]

    def compute_if_not_computed(self):
        if self.result is None:
            self.compute()
            
    def mass_marginal_plot(self, ax=None, aspect=0.5, log_lower=-15, color=None, label=None, alpha=0.3):
        self.compute_if_not_computed()
        mass_name = [m.name for m in self.mass_grid.grid_list if "mass_1" in m.name][0]
        mass_ratio_name = [m.name for m in self.mass_grid.grid_list if (("mass" in m.name) and ("ratio" in m.name))][0]
        
        mass_1_grid_obj = [m for m in self.mass_grid.grid_list if "mass_1" in m.name][0]
        mass_ratio_grid_obj = [m for m in self.mass_grid.grid_list if (("mass" in m.name) and ("ratio" in m.name))][0]
        
        mass_1_grid = self.mass_grid.grid_1ds[mass_name]
        mass_ratio_grid = self.mass_grid.grid_1ds[mass_ratio_name]
        
        mass_x = mass_1_grid
        mass_y_all = jax.scipy.integrate.trapezoid(self.result, mass_ratio_grid, axis=1)
        mass_y_median = jnp.quantile(mass_y_all, 0.5, axis=0)
        mass_y_lower = jnp.quantile(mass_y_all, self.conf[0], axis=0)
        mass_y_upper = jnp.quantile(mass_y_all, self.conf[1], axis=0)
        
        highest, lowest = jnp.log10(mass_y_upper.max()),  max(jnp.log10(mass_y_lower.min()), log_lower)
        high_x, low_x = mass_x.max(), mass_x.min()
        
        import matplotlib.pyplot as plt
        if ax is None:
            fig,ax = plt.subplots(1)
        ax.plot(mass_x, mass_y_median, color=color, label=label)
        ax.fill_between(mass_x, mass_y_lower, mass_y_upper, color=color, alpha=alpha)
        ax.set_yscale("log")
        ax.set_xlabel(mass_1_grid_obj.latex_name)
        if self.rate:
            ax.set_ylabel(r"$\frac{dN}{dm_1}$")
        else:
            ax.set_ylabel(r"$p(m_1 | \Lambda)$")
        ax.grid(True, which='major', linestyle='dotted', linewidth='0.5', color='gray', alpha=0.5)
        ax.set_xlim(mass_x.min(), mass_x.max())
        ax.set_ylim(bottom=10**(lowest))
        new_aspect = aspect*(high_x - low_x)/(highest-lowest)
        if (new_aspect > 0) and not (np.isinf(new_aspect)):
            ax.set_aspect(aspect*(high_x - low_x)/(highest-lowest))

        return ax
    
    def mass_ratio_marginal_plot(self, ax=None, aspect=0.5, log_lower=-15, color=None, label=None, alpha=0.3):
        self.compute_if_not_computed()
        mass_name = [m.name for m in self.mass_grid.grid_list if "mass_1" in m.name][0]
        mass_ratio_name = [m.name for m in self.mass_grid.grid_list if (("mass" in m.name) and ("ratio" in m.name))][0]
        
        mass_1_grid_obj = [m for m in self.mass_grid.grid_list if "mass_1" in m.name][0]
        mass_ratio_grid_obj = [m for m in self.mass_grid.grid_list if (("mass" in m.name) and ("ratio" in m.name))][0]
        
        mass_1_grid = self.mass_grid.grid_1ds[mass_name]
        mass_ratio_grid = self.mass_grid.grid_1ds[mass_ratio_name]
        
        mass_x = mass_ratio_grid
        mass_ratio_y_all = jax.scipy.integrate.trapezoid(self.result, mass_1_grid, axis=2)
        mass_ratio_y_median = jnp.quantile(mass_ratio_y_all, 0.5, axis=0)
        mass_ratio_y_lower = jnp.quantile(mass_ratio_y_all, self.conf[0], axis=0)
        mass_ratio_y_upper = jnp.quantile(mass_ratio_y_all, self.conf[1], axis=0)
        
        highest, lowest = jnp.log10(mass_ratio_y_upper.max()),  max(jnp.log10(mass_ratio_y_lower.min()), log_lower)
        high_x, low_x = mass_x.max(), mass_x.min()
        
        import matplotlib.pyplot as plt
        if ax is None:
            fig,ax = plt.subplots(1)
        ax.plot(mass_x, mass_ratio_y_median, color=color, label=label)
        ax.fill_between(mass_x, mass_ratio_y_lower, mass_ratio_y_upper, color=color, alpha=alpha)
        ax.set_xlabel(mass_ratio_grid_obj.latex_name)
        if self.rate:
            ax.set_ylabel(r"$\frac{dN}{dq}$")
        else:
            ax.set_ylabel(r"$p(q | \Lambda)$")
        ax.grid(True, which='major', linestyle='dotted', linewidth='0.5', color='gray', alpha=0.5)
        ax.set_xlim(mass_x.min(), mass_x.max())
        ax.set_yscale("log")
        ax.set_ylim(bottom=10**(log_lower))
        new_aspect = aspect*(high_x - low_x)/(highest-lowest)
        if (new_aspect > 0) and not (np.isinf(new_aspect)):
            ax.set_aspect(aspect*(high_x - low_x)/(highest-lowest))
        #ax.set_aspect(aspect*(high_x - low_x)/(highest-lowest))
        #ax.set_aspect(aspect*(1)/(highest-lowest))
        
        return ax


#### This is a direct rip from gwpop so I can easily test against it

import inspect
import json
import os
import sys
from importlib import import_module

import matplotlib

#matplotlib.use("agg")  # noqa

import dill
import numpy as np
import pandas as pd
from bilby import run_sampler
from bilby.core.prior import Constraint, LogUniform, ConditionalPriorDict
from bilby.core.utils import (
    infer_args_from_function_except_n_args,
    logger,
    decode_bilby_json,
)
from bilby.hyper.model import Model
from bilby_pipe.utils import convert_string_to_dict
from gwpopulation.backend import set_backend
from gwpopulation.conversions import convert_to_beta_parameters as convert_to_beta_parameters_gwpop
from gwpopulation.hyperpe import HyperparameterLikelihood, RateLikelihood
from gwpopulation.models.mass import (
    BrokenPowerLawPeakSmoothedMassDistribution,
    BrokenPowerLawSmoothedMassDistribution,
    MultiPeakSmoothedMassDistribution,
    SinglePeakSmoothedMassDistribution,
    two_component_primary_mass_ratio,
)
from gwpopulation.models.spin import (
    iid_spin,
    iid_spin_magnitude_beta,
    iid_spin_orientation_gaussian_isotropic,
    independent_spin_magnitude_beta,
    independent_spin_orientation_gaussian_isotropic,
)
from gwpopulation.utils import to_numpy
from scipy.stats import gamma
from tqdm.auto import trange

from gwpopulation_pipe import vt_helper
from gwpopulation_pipe.parser import create_parser as create_main_parser
from gwpopulation_pipe.utils import (
    get_path_or_local,
    prior_conversion,
    KNOWN_ARGUMENTS,
    MinimumEffectiveSamplesLikelihood,
)


MODEL_MAP = {
    "two_component_primary_mass_ratio": two_component_primary_mass_ratio,
    "iid_spin": iid_spin,
    "iid_spin_magnitude": iid_spin_magnitude_beta,
    "ind_spin_magnitude": independent_spin_magnitude_beta,
    "iid_spin_orientation": iid_spin_orientation_gaussian_isotropic,
    "two_comp_iid_spin_orientation": iid_spin_orientation_gaussian_isotropic,
    "ind_spin_orientation": independent_spin_orientation_gaussian_isotropic,
    "SmoothedMassDistribution": SinglePeakSmoothedMassDistribution,
    "SinglePeakSmoothedMassDistribution": SinglePeakSmoothedMassDistribution,
    "BrokenPowerLawSmoothedMassDistribution": BrokenPowerLawSmoothedMassDistribution,
    "MultiPeakSmoothedMassDistribution": MultiPeakSmoothedMassDistribution,
    "BrokenPowerLawPeakSmoothedMassDistribution": BrokenPowerLawPeakSmoothedMassDistribution,
}



from dataclasses import dataclass, field
from typing import Union, List, Optional


#DEFAULT_MASS = {'mass' : 'SmoothedMassDistribution', 
#                'redshift' : 'gwpopulation.models.redshift.PowerLawRedshift', 
#               'tilt' : 'iid_spin', 
#                'mag' : 'iid_spin_orientation'}

DEFAULT_MASS = dict(
            mass="SmoothedMassDistribution",#"two_component_primary_mass_ratio",
            mag="iid_spin_magnitude",
            tilt="iid_spin_orientation",
            redshift="gwpopulation.models.redshift.PowerLawRedshift",
        )


@dataclass
class GWPopLoader:
    prior_file : str
    vt_file : str
    posterior_file : Optional[str] = None
    backend : str  = 'jax'
    conversion_function = None
    rate : bool = False
    models : List[str] = field(default_factory=(lambda : DEFAULT_MASS.copy()))
    minimum_mass: float = 2
    maximum_mass : float = 100
    max_redshift : float = 1.9
    vt_models : List[str] = field(default_factory=(lambda : DEFAULT_MASS.copy()))
    vt_ifar_threshold : float = 1.0
    vt_snr_threshold : float = 11.0
    vt_function : str = "injection_resampling_vt"
    enforce_minimum_neffective_per_event : bool = False
    samples_per_posterior : int = 1000
    
    def __post_init__(self):
        self.posteriors = None
        self.prior = None
        self.selection = None
        self.likelihood = None


    def _load_model(self, model):
        args = self
        if model[-5:] == ".json":
            model = get_path_or_local(model)
            with open(model, "r") as ff:
                json_model = json.load(ff, object_hook=decode_bilby_json)
            try:
                cls = getattr(import_module(json_model["module"]), json_model["class"])
                _model = cls(**json_model.get("kwargs", dict()))
                logger.info(f"Using {cls} from {json_model['module']}.")
            except KeyError:
                logger.error(f"Failed to load {model} from json file.")
                raise
        elif "." in model:
            split_model = model.split(".")
            module = ".".join(split_model[:-1])
            function = split_model[-1]
            _model = getattr(import_module(module), function)
            logger.info(f"Using {function} from {module}.")
        elif model in MODEL_MAP:
            _model = MODEL_MAP[model]
            logger.info(f"Using {model}.")
        else:
            raise ValueError(f"Model {model} not found.")
        if inspect.isclass(_model):
            if "redshift" in model.lower():
                kwargs = dict(z_max=args.max_redshift)
            elif "mass" in model.lower():
                kwargs = dict(mmin=args.minimum_mass, mmax=args.maximum_mass)
            else:
                kwargs = dict()
            try:
                _model = _model(**kwargs)
                logger.info(f"Created {model} with arguments {kwargs}")
            except TypeError:
                logger.warning(f"Failed to instantiate {model} with arguments {kwargs}")
                _model = _model()
        return _model
    
    
    def load_prior(self):
        filename = get_path_or_local(self.prior_file)
        hyper_prior = ConditionalPriorDict(filename=filename)
        hyper_prior.conversion_function = convert_to_beta_parameters_gwpop
        if self.rate:
            hyper_prior["rate"] = LogUniform(
                minimum=1e-1,
                maximum=1e3,
                name="rate",
                latex_label="$R$",
                boundary="reflective",
            )
        self.prior = hyper_prior
        return hyper_prior
    
    def load_model(self):
        args = self
        if args.models is None:
            args.models = dict(
                mass="two_component_primary_mass_ratio",
                mag="iid_spin_magnitude",
                tilt="iid_spin_orientation",
                redshift="gwpopulation.models.redshift.PowerLawRedshift",
            )
        if args.backend == "jax":
            from gwpopulation.experimental.jax import NonCachingModel

            cls = NonCachingModel
        else:
            cls = Model
        model = cls([self._load_model(model) for model in args.models.values()])
        self.model = model
        return model
    
    def load_vt(self):
        args = self
        if args.vt_function == "" or args.vt_file == "None":
            result = vt_helper.dummy_selection
            self.selection = result
            return result
        if args.backend == "jax":
            from gwpopulation.experimental.jax import NonCachingModel

            cls = NonCachingModel
        else:
            cls = Model
        vt_model = cls([self._load_model(model) for model in args.vt_models.values()])
        try:
            vt_func = getattr(vt_helper, args.vt_function)
            result = vt_func(
                args.vt_file,
                model=vt_model,
                ifar_threshold=args.vt_ifar_threshold,
                snr_threshold=args.vt_snr_threshold,
            )
            self.selection = result
            return result
        except AttributeError:
            result = vt_helper.injection_resampling_vt(
                vt_file=args.vt_file,
                model=vt_model,
                ifar_threshold=args.vt_ifar_threshold,
                snr_threshold=args.vt_snr_threshold)
            self.selection = result
            return result

    def load_posteriors(self, posterior_file=None):
        import numpy as np
        import pandas as pd
        file = posterior_file or self.posterior_file
        posteriors = np.load(self.posterior_file, allow_pickle=True)
        self.posteriors = posteriors

    def create_likelihood(self, posteriors=None):
        args = self
        if posteriors is None:
            if self.posteriors is None:
                self.load_posteriors()
            posteriors = self.posteriors
        if args.rate:
            if args.enforce_minimum_neffective_per_event:
                raise ValueError(
                    "No likelihood available to enforce convergence of Monte Carlo integrals "
                    "while sampling over rate."
                )
            likelihood_class = RateLikelihood
        elif args.enforce_minimum_neffective_per_event:
            likelihood_class = MinimumEffectiveSamplesLikelihood
        else:
            likelihood_class = HyperparameterLikelihood
        likelihood = likelihood_class(
            posteriors,
            self.model,
            conversion_function=convert_to_beta_parameters_gwpop,
            selection_function=self.selection,
            max_samples=self.samples_per_posterior,
            cupy=False
        )
        self.likelihood = likelihood
        return likelihood


from ..generic import *


class TruncatedGaussian1DAnalyticSpin(TruncatedGaussian1DAnalytic, SpinPopulationModel):
    pass

class TruncatedGaussian2DAnalyticSpin(TruncatedGaussian2DAnalytic, SpinPopulationModel):
    pass
